# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Serializers for datasets."""
import textwrap

from renku.core.models.json import dumps

from .tabulate import tabulate


def tabular(client, datasets, *, columns=None):
    """Format datasets with a tabular output."""
    if not columns:
        columns = "id,date_created,name,creators,tags,version"

    _create_dataset_short_description(datasets)

    return tabulate(collection=datasets, columns=columns, columns_mapping=DATASETS_COLUMNS)


def _create_dataset_short_description(datasets):
    for dataset in datasets:
        lines = textwrap.wrap(dataset.description, width=64, max_lines=5) if dataset.description else []
        dataset.short_description = "\n".join(lines)


def jsonld(client, datasets, **kwargs):
    """Format datasets as JSON-LD."""
    data = [dataset.as_jsonld() for dataset in datasets]
    return dumps(data, indent=2)


DATASETS_FORMATS = {
    "tabular": tabular,
    "json-ld": jsonld,
}
"""Valid formatting options."""

DATASETS_COLUMNS = {
    "id": ("identifier", "id"),
    "created": ("date_created", None),
    "date_created": ("date_created", None),
    "short_name": ("name", None),
    "name": ("name", None),
    "creators": ("creators_csv", "creators"),
    "creators_full": ("creators_full_csv", "creators"),
    "tags": ("tags_csv", "tags"),
    "version": ("version", None),
    "title": ("title", "title"),
    "keywords": ("keywords_csv", "keywords"),
    "description": ("short_description", "description"),
}



# -*- coding: utf-8 -*-
#
# Copyright 2018-2020- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Represent provenance agents."""

import os
import pathlib
import re
import urllib
import uuid
from urllib.parse import quote

import attr
from attr.validators import instance_of
from calamus.schema import JsonLDSchema
from marshmallow import EXCLUDE

from renku.core.models.calamus import StringList, fields, prov, rdfs, schema, wfprov
from renku.core.models.git import get_user_info
from renku.version import __version__, version_url


@attr.s(slots=True,)
class Person:
    """Represent a person."""

    client = attr.ib(default=None, kw_only=True)

    name = attr.ib(kw_only=True, validator=instance_of(str))
    email = attr.ib(default=None, kw_only=True)
    label = attr.ib(kw_only=True)
    affiliation = attr.ib(default=None, kw_only=True,)
    alternate_name = attr.ib(default=None, kw_only=True,)
    _id = attr.ib(default=None, kw_only=True)

    def default_id(self):
        """Set the default id."""
        return generate_person_id(email=self.email, client=self.client, full_identity=self.full_identity)

    @email.validator
    def check_email(self, attribute, value):
        """Check that the email is valid."""
        if self.email and not (isinstance(value, str) and re.match(r"[^@]+@[^@]+\.[^@]+", value)):
            raise ValueError("Email address is invalid.")

    @label.default
    def default_label(self):
        """Set the default label."""
        return self.name

    @classmethod
    def from_commit(cls, commit):
        """Create an instance from a Git commit."""
        return cls(name=commit.author.name, email=commit.author.email,)

    @property
    def short_name(self):
        """Gives full name in short form."""
        names = self.name.split()
        if len(names) == 1:
            return self.name

        last_name = names[-1]
        initials = [name[0] for name in names]
        initials.pop()

        return "{0}.{1}".format(".".join(initials), last_name)

    @property
    def full_identity(self):
        """Return name, email, and affiliation."""
        email = f" <{self.email}>" if self.email else ""
        affiliation = f" [{self.affiliation}]" if self.affiliation else ""
        return f"{self.name}{email}{affiliation}"

    @classmethod
    def from_git(cls, git):
        """Create an instance from a Git repo."""
        name, email = get_user_info(git)
        return cls(name=name, email=email)

    @classmethod
    def from_string(cls, string):
        """Create an instance from a 'Name <email>' string."""
        regex_pattern = r"([^<>\[\]]*)" r"(?:<{1}\s*(\S+@\S+\.\S+){0,1}\s*>{1}){0,1}\s*" r"(?:\[{1}(.*)\]{1}){0,1}"
        name, email, affiliation = re.search(regex_pattern, string).groups()
        if name:
            name = name.strip()
        if affiliation:
            affiliation = affiliation.strip()
        affiliation = affiliation or None

        return cls(name=name, email=email, affiliation=affiliation)

    @classmethod
    def from_dict(cls, obj):
        """Create and instance from a dictionary."""
        return cls(**obj)

    @classmethod
    def from_jsonld(cls, data):
        """Create an instance from JSON-LD data."""
        if isinstance(data, cls):
            return data
        if not isinstance(data, dict):
            raise ValueError(data)

        return PersonSchema().load(data)

    def __attrs_post_init__(self):
        """Finish object initialization."""
        # handle the case where ids were improperly set
        if self._id == "mailto:None" or not self._id or self._id.startswith("_:"):
            self._id = self.default_id()

        if self.label is None:
            self.label = self.default_label()


class PersonSchema(JsonLDSchema):
    """Person schema."""

    class Meta:
        """Meta class."""

        rdf_type = [prov.Person, schema.Person]
        model = Person
        unknown = EXCLUDE

    name = StringList(schema.name, missing=None)
    email = fields.String(schema.email, missing=None)
    label = StringList(rdfs.label, missing=None)
    affiliation = StringList(schema.affiliation, missing=None)
    alternate_name = StringList(schema.alternateName, missing=None)
    _id = fields.Id(init_name="id")


@attr.s(
    frozen=True, slots=True,
)
class SoftwareAgent:
    """Represent executed software."""

    label = attr.ib(kw_only=True)

    _id = attr.ib(kw_only=True)

    @classmethod
    def from_commit(cls, commit):
        """Create an instance from a Git commit."""
        author = Person.from_commit(commit)
        if commit.author != commit.committer:
            return cls(label=commit.committer.name, id=commit.committer.email)
        return author

    @classmethod
    def from_jsonld(cls, data):
        """Create an instance from JSON-LD data."""
        if isinstance(data, cls):
            return data
        if not isinstance(data, dict):
            raise ValueError(data)

        return SoftwareAgentSchema().load(data)

    def as_jsonld(self):
        """Create JSON-LD."""
        return SoftwareAgentSchema().dump(self)


# set up the default agent

renku_agent = SoftwareAgent(label="renku {0}".format(__version__), id=version_url)


def generate_person_id(client, email, full_identity):
    """Generate Person default id."""
    if email:
        return "mailto:{email}".format(email=email)

    host = "localhost"
    if client:
        host = client.remote.get("host") or host
    host = os.environ.get("RENKU_DOMAIN") or host

    id_ = full_identity or str(uuid.uuid4())

    return urllib.parse.urljoin(
        "https://{host}".format(host=host), pathlib.posixpath.join("/persons", quote(id_, safe=""))
    )


class SoftwareAgentSchema(JsonLDSchema):
    """SoftwareAgent schema."""

    class Meta:
        """Meta class."""

        rdf_type = [prov.SoftwareAgent, wfprov.WorkflowEngine]
        model = SoftwareAgent
        unknown = EXCLUDE

    label = fields.String(rdfs.label)
    _id = fields.Id(init_name="id")



# -*- coding: utf-8 -*-
#
# Copyright 2017-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Storage command tests."""

import os
import subprocess
from pathlib import Path

from renku.cli import cli


def test_lfs_storage_clean_no_remote(runner, project, client):
    """Test ``renku storage clean`` command with no remote set."""
    with (client.path / "tracked").open("w") as fp:
        fp.write("tracked file")
    client.repo.git.add("*")
    client.repo.index.commit("tracked file")

    result = runner.invoke(cli, ["storage", "clean", "tracked"], catch_exceptions=False)
    assert 1 == result.exit_code
    assert "No git remote is configured for" in result.output


def test_lfs_storage_clean(runner, project, client_with_remote):
    """Test ``renku storage clean`` command."""
    client = client_with_remote["client"]

    with (client.path / "tracked").open("w") as fp:
        fp.write("tracked file")
    client.repo.git.add("*")
    client.repo.index.commit("tracked file")

    result = runner.invoke(cli, ["storage", "clean", "tracked"], catch_exceptions=False)
    assert 0 == result.exit_code
    assert "These paths were ignored as they are not tracked" in result.output

    subprocess.call(["git", "lfs", "track", "tracked"])
    client.repo.git.add("*")
    client.repo.index.commit("Tracked in lfs")
    client.repo.git.push("--no-verify", "origin")

    with (client.path / "tracked").open("r") as fp:
        assert "tracked file" in fp.read()

    assert "tracked" in Path(".gitattributes").read_text()

    lfs_objects = []
    for _, _, files in os.walk(str(client.path / ".git" / "lfs" / "objects")):
        lfs_objects.extend(files)

    assert 1 == len(lfs_objects)

    result = runner.invoke(cli, ["storage", "clean", "tracked"], catch_exceptions=False)
    assert 0 == result.exit_code

    assert "version https://git-lfs.github.com/spec/v1" in (client.path / "tracked").read_text()

    lfs_objects = []
    for _, _, files in os.walk(str(client.path / ".git" / "lfs" / "objects")):
        lfs_objects.extend(files)

    assert 0 == len(lfs_objects)

    # already clean file should be ignored on clean
    result = runner.invoke(cli, ["storage", "clean", "tracked"], catch_exceptions=False)
    assert 0 == result.exit_code


def test_lfs_storage_unpushed_clean(runner, project, client_with_remote):
    """Test ``renku storage clean`` command for unpushed files."""
    client = client_with_remote["client"]

    with (client.path / "tracked").open("w") as fp:
        fp.write("tracked file")
    subprocess.call(["git", "lfs", "track", "tracked"])
    client.repo.git.add("*")
    client.repo.index.commit("tracked file")

    result = runner.invoke(cli, ["storage", "clean", "tracked"], catch_exceptions=False)

    assert 0 == result.exit_code
    assert "These paths were ignored as they are not pushed" in result.output



# -*- coding: utf-8 -*-
#
# Copyright 2017-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for Run API."""

from pathlib import Path

from renku.api import Input, Output, Parameter, Project
from renku.core.models.cwl import command_line_tool
from renku.core.models.cwl.command_line_tool import read_indirect_parameters


def test_indirect_inputs(client):
    """Test defining indirect inputs."""
    path_1 = "/some/absolute/path"
    path_2 = "relative/path"
    path_3 = "a/path with white-spaces/"

    input_1 = Input(path_1)

    with Project() as project:
        input_2 = Input(path_2)

    input_3 = Input(path_3)

    assert Path(path_1) == input_1.path
    assert Path(path_2) == input_2.path
    assert Path(path_3) == input_3.path

    content = command_line_tool.get_indirect_inputs_path(project.path).read_text()

    assert {path_1, path_2, path_3} == {line for line in content.split("\n") if line}


def test_indirect_outputs(client):
    """Test defining indirect outputs."""
    path_1 = "/some/absolute/path"
    path_2 = "relative/path"
    path_3 = "a/path with white-spaces/"

    input_1 = Output(path_1)

    with Project() as project:
        input_2 = Output(path_2)

    input_3 = Output(path_3)

    assert Path(path_1) == input_1.path
    assert Path(path_2) == input_2.path
    assert Path(path_3) == input_3.path

    content = command_line_tool.get_indirect_outputs_path(project.path).read_text()

    assert {path_1, path_2, path_3} == {line for line in content.split("\n") if line}


def test_indirect_inputs_outputs(client):
    """Test defining indirect inputs and outputs together."""
    path_1 = "/some/absolute/path"
    path_2 = "relative/path"

    input_1 = Input(path_1)
    output_2 = Output(path_2)

    assert Path(path_1) == input_1.path
    assert Path(path_2) == output_2.path

    assert path_1 == command_line_tool.get_indirect_inputs_path(client.path).read_text().strip()
    assert path_2 == command_line_tool.get_indirect_outputs_path(client.path).read_text().strip()


def test_open_inputs(client):
    """Test inputs can be passed to open function."""
    with open(Input("input.txt"), "w") as f:
        f.write("some data")

    assert "some data" == (client.path / "input.txt").read_text()


def test_open_outputs(client):
    """Test outputs can be passed to open function."""
    with open(Output("output.txt"), "w") as f:
        f.write("some data")

    assert "some data" == (client.path / "output.txt").read_text()


def test_parameters(client):
    """Test defining parameters."""
    p1 = Parameter("parameter 1", 42)

    with Project():
        p2 = Parameter("param-2", "42")

    p3 = Parameter(" parameter 3 ", 42.42)

    assert (42, "42", 42.42) == (p1, p2, p3)

    data = read_indirect_parameters(client.path)

    assert {"parameter 1", "param-2", " parameter 3 "} == set(data.keys())
    assert {42, "42", 42.42} == set(data.values())



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service cache upload files controller."""
import os
import shutil
from pathlib import Path

import patoolib
from patoolib.util import PatoolError

from renku.core.errors import RenkuException
from renku.service.config import CACHE_UPLOADS_PATH, SUPPORTED_ARCHIVES
from renku.service.controllers.api.abstract import ServiceCtrl
from renku.service.controllers.api.mixins import ReadOperationMixin
from renku.service.serializers.cache import FileUploadRequest, FileUploadResponseRPC, extract_file
from renku.service.views import result_response


class UploadFilesCtrl(ServiceCtrl, ReadOperationMixin):
    """Controller for upload files endpoint."""

    REQUEST_SERIALIZER = FileUploadRequest()
    RESPONSE_SERIALIZER = FileUploadResponseRPC()

    def __init__(self, cache, user_data, flask_request):
        """Construct controller."""
        self.file = extract_file(flask_request)

        self.response_builder = {
            "file_name": self.file.filename,
            "content_type": self.file.content_type,
            "is_archive": self.file.content_type in SUPPORTED_ARCHIVES,
        }
        self.response_builder.update(UploadFilesCtrl.REQUEST_SERIALIZER.load(flask_request.args))

        super(UploadFilesCtrl, self).__init__(cache, user_data, {})

    @property
    def context(self):
        """Controller operation context."""
        return self.response_builder

    def process_file(self):
        """Process uploaded file."""
        user_cache_dir = CACHE_UPLOADS_PATH / self.user.user_id
        user_cache_dir.mkdir(exist_ok=True)

        file_path = user_cache_dir / self.file.filename
        if file_path.exists():
            if self.response_builder.get("override_existing", False):
                file_path.unlink()
            else:
                raise RenkuException("file exists")

        self.file.save(str(file_path))

        files = []
        if self.response_builder["unpack_archive"] and self.response_builder["is_archive"]:
            unpack_dir = "{0}.unpacked".format(file_path.name)
            temp_dir = file_path.parent / Path(unpack_dir)
            if temp_dir.exists():
                shutil.rmtree(temp_dir)
            temp_dir.mkdir(exist_ok=True)

            try:
                patoolib.extract_archive(str(file_path), outdir=str(temp_dir))
            except PatoolError:
                return RenkuException("unable to unpack archive")

            for file_ in temp_dir.glob("**/*"):
                relative_path = file_.relative_to(user_cache_dir)

                file_obj = {
                    "file_name": file_.name,
                    "file_size": os.stat(file_).st_size,
                    "relative_path": str(relative_path),
                    "is_dir": file_.is_dir(),
                }

                files.append(file_obj)

        else:
            relative_path = file_path.relative_to(CACHE_UPLOADS_PATH / self.user.user_id)

            self.response_builder["file_size"] = os.stat(file_path).st_size
            self.response_builder["relative_path"] = str(relative_path)
            self.response_builder["is_dir"] = file_path.is_dir()

            files.append(self.response_builder)

        files = self.cache.set_files(self.user, files)
        return {"files": files}

    def renku_op(self):
        """Renku operation for the controller."""
        # NOTE: We leave it empty since it does not execute renku operation.
        pass

    def to_response(self):
        """Execute controller flow and serialize to service response."""
        return result_response(UploadFilesCtrl.RESPONSE_SERIALIZER, self.process_file())



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service migrations check controller."""

from renku.core.commands.migrate import migrations_check, migrations_versions
from renku.service.controllers.api.abstract import ServiceCtrl
from renku.service.controllers.api.mixins import ReadOperationMixin
from renku.service.serializers.cache import ProjectMigrationCheckRequest, ProjectMigrationCheckResponseRPC
from renku.service.views import result_response


class MigrationsCheckCtrl(ServiceCtrl, ReadOperationMixin):
    """Controller for migrations check endpoint."""

    REQUEST_SERIALIZER = ProjectMigrationCheckRequest()
    RESPONSE_SERIALIZER = ProjectMigrationCheckResponseRPC()

    def __init__(self, cache, user_data, request_data):
        """Construct migration check controller."""
        self.ctx = MigrationsCheckCtrl.REQUEST_SERIALIZER.load(request_data)
        super(MigrationsCheckCtrl, self).__init__(cache, user_data, request_data)

    @property
    def context(self):
        """Controller operation context."""
        return self.ctx

    def renku_op(self):
        """Renku operation for the controller."""
        latest_version, project_version = migrations_versions()
        (
            migration_required,
            project_supported,
            template_update_possible,
            current_template_version,
            latest_template_version,
            automated_update,
            docker_update_possible,
        ) = migrations_check()

        return {
            "migration_required": migration_required,
            "template_update_possible": template_update_possible,
            "current_template_version": current_template_version,
            "latest_template_version": latest_template_version,
            "automated_template_update": automated_update,
            "docker_update_possible": docker_update_possible,
            "project_supported": project_supported,
            "project_version": project_version,
            "latest_version": latest_version,
        }

    def to_response(self):
        """Execute controller flow and serialize to service response."""
        return result_response(MigrationsCheckCtrl.RESPONSE_SERIALIZER, self.execute_op())



# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service version view."""
from flask import Blueprint

from renku.service.config import SERVICE_PREFIX
from renku.service.controllers.version import VersionCtrl
from renku.service.views.decorators import handle_validation_except, header_doc

VERSION_BLUEPRINT_TAG = "version"
version_blueprint = Blueprint("version", __name__, url_prefix=SERVICE_PREFIX)


@header_doc(description="Display version of the service.", tags=(VERSION_BLUEPRINT_TAG,))
@version_blueprint.route(
    "/version", methods=["GET"], provide_automatic_options=False,
)
@handle_validation_except
def version():
    """Version view."""
    return VersionCtrl().to_response()



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service project remote abstraction tests."""
import pytest
from marshmallow import ValidationError

import renku
from renku.core.commands.migrate import migrations_check, migrations_versions
from renku.service.controllers.utils.remote_project import RemoteProject


def test_project_metadata_remote():
    """Check path construction for remote project metadata path."""
    user_data = {
        "fullname": "testing user",
        "email": "testing@user.com",
        "token": "123",
    }
    request_data = {"git_url": "https://dev.renku.ch/gitlab/contact/import-me"}
    ctrl = RemoteProject(user_data, request_data)
    path = ctrl.remote_url

    assert path
    assert "https" == path.scheme
    assert "oauth2:123@dev.renku.ch" == path.netloc
    assert "/gitlab/contact/import-me" == path.path
    assert "" == path.params
    assert "" == path.query
    assert "" == path.fragment


def test_project_metadata_custom_remote():
    """Check path construction for remote project metadata path."""
    user_data = {
        "fullname": "testing user",
        "email": "testing@user.com",
        "token": "123",
    }

    request_data = {"git_url": "https://dev.renku.ch/gitlab/contact/import-me", "ref": "my-branch"}

    ctrl = RemoteProject(user_data, request_data)
    ref = ctrl.ctx["ref"]

    assert request_data["ref"] == ref


def test_project_metadata_remote_err():
    """Check exception raised during path construction for remote project metadata path."""
    user_data = {
        "fullname": "testing user",
        "email": "testing@user.com",
        "token": "123",
    }
    request_data = {"git_url": "/dev.renku.ch/gitlab/contact/import-me"}

    with pytest.raises(ValidationError):
        RemoteProject(user_data, request_data)

    request_data["git_url"] = "httpz://dev.renku.ch/gitlab/contact/import-me"

    with pytest.raises(ValidationError):
        RemoteProject(user_data, request_data)


def test_remote_project_context():
    """Check remote project context manager."""
    user_data = {
        "fullname": "testing user",
        "email": "testing@user.com",
        "token": "123",
    }
    request_data = {"git_url": "https://dev.renku.ch/gitlab/contact/import-me"}
    ctrl = RemoteProject(user_data, request_data)

    with ctrl.remote() as project_path:
        assert project_path
        latest_version, project_version = migrations_versions()
        assert renku.__version__ == latest_version
        assert "pre-0.11.0" == project_version

        (
            migration_required,
            project_supported,
            template_update_possible,
            current_template_version,
            latest_template_version,
            automated_update_possible,
            docker_update_possible,
        ) = migrations_check()
        assert migration_required is True
        assert template_update_possible is False
        assert current_template_version is None
        assert latest_template_version is None
        assert automated_update_possible is False
        assert docker_update_possible is False
        assert project_supported is True



# -*- coding: utf-8 -*-
#
# Copyright 2018-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Represents a workflow template."""

import os
import pathlib
import urllib.parse
import uuid
from bisect import bisect
from copy import copy
from functools import total_ordering
from pathlib import Path

import attr
from marshmallow import EXCLUDE

from renku.core.models.calamus import JsonLDSchema, Nested, fields, prov, renku
from renku.core.models.cwl.types import PATH_OBJECTS
from renku.core.models.entities import Collection, CommitMixin, CommitMixinSchema, Entity
from renku.core.models.workflow.parameters import (
    CommandArgument,
    CommandArgumentSchema,
    CommandInput,
    CommandInputSchema,
    CommandOutput,
    CommandOutputSchema,
    MappedIOStream,
    RunParameter,
    RunParameterSchema,
)


def _entity_from_path(client, path, commit):
    """Gets the entity associated with a path."""
    client, commit, path = client.resolve_in_submodules(
        client.find_previous_commit(path, revision=commit.hexsha), path,
    )

    entity_cls = Entity
    if (client.path / path).is_dir():
        entity_cls = Collection

    if str(path).startswith(os.path.join(client.renku_home, client.DATASETS)):
        return client.load_dataset_from_path(path, commit=commit)
    else:
        return entity_cls(commit=commit, client=client, path=str(path),)


def _convert_cmd_binding(binding, client, commit):
    """Convert a cwl argument to ``CommandArgument``."""

    base_id = Run.generate_id(client)

    id_ = CommandArgument.generate_id(base_id, binding.position)

    return CommandArgument(id=id_, position=binding.position, value=binding.valueFrom)


def _convert_cmd_input(input, client, commit, run_id):
    """Convert a cwl input to ``CommandInput``."""
    val = input.default

    if isinstance(val, list):
        val = input.inputBinding.itemSeparator.join(val)

    if input.type in PATH_OBJECTS and input.default:
        if input.inputBinding:
            prefix = input.inputBinding.prefix
            if prefix and input.inputBinding.separate:
                prefix += " "
            return CommandInput(
                id=CommandInput.generate_id(run_id, input.inputBinding.position),
                position=input.inputBinding.position,
                prefix=prefix,
                consumes=_entity_from_path(client, input.default.path, commit),
            )
        else:
            return CommandInput(
                id=CommandInput.generate_id(run_id, "stdin" if input.id == "input_stdin" else None),
                consumes=_entity_from_path(client, input.default.path, commit),
                mapped_to=MappedIOStream(client=client, stream_type="stdin") if input.id == "input_stdin" else None,
            )
    else:
        prefix = input.inputBinding.prefix
        if prefix and input.inputBinding.separate:
            prefix += " "
        return CommandArgument(
            id=CommandArgument.generate_id(run_id, input.inputBinding.position),
            position=input.inputBinding.position,
            value=val,
            prefix=prefix,
        )


def _convert_cmd_output(output, factory, client, commit, run_id):
    """Convert a cwl output to ``CommandOutput``."""
    path = None
    mapped = None
    input_prefix = "$(inputs."
    position = None
    prefix = None
    input_to_remove = None
    create_folder = False

    if output.outputBinding:
        if output.outputBinding.glob.startswith(input_prefix):
            input_id = output.outputBinding.glob[len(input_prefix) : -1]
            inp = next(i for i in factory.inputs if i.id == input_id)
            path = inp.default
            position = inp.inputBinding.position
            prefix = inp.inputBinding.prefix
            if prefix and inp.inputBinding.separate:
                prefix += " "
            input_to_remove = inp
        else:
            path = output.outputBinding.glob

    if output.type in MappedIOStream.STREAMS:
        path = getattr(factory, output.type)
        mapped = MappedIOStream(client=client, stream_type=output.type)

    if ((client.path / path).is_dir() and path in factory.existing_directories) or (
        not (client.path / path).is_dir() and str(Path(path).parent) in factory.existing_directories
    ):
        create_folder = True

    return (
        CommandOutput(
            id=CommandOutput.generate_id(run_id, position),
            produces=_entity_from_path(client, path, commit),
            mapped_to=mapped,
            position=position,
            prefix=prefix,
            create_folder=create_folder,
        ),
        input_to_remove,
    )


def _convert_run_parameter(parameter, run_id):
    """Convert a cwl run parameters to ``RunParameter``."""
    id_ = RunParameter.generate_id(run_id=run_id, name=parameter.name)
    return RunParameter(id=id_, name=parameter.name, value=parameter.value)


@total_ordering
@attr.s(cmp=False,)
class Run(CommitMixin):
    """Represents a `renku run` execution template."""

    command = attr.ib(default=None, type=str, kw_only=True,)

    successcodes = attr.ib(kw_only=True, type=list, factory=list)

    subprocesses = attr.ib(kw_only=True, factory=list)

    arguments = attr.ib(kw_only=True, factory=list)

    inputs = attr.ib(kw_only=True, factory=list)

    outputs = attr.ib(kw_only=True, factory=list)

    run_parameters = attr.ib(kw_only=True, factory=list)

    _activity = attr.ib(kw_only=True, default=None)

    @staticmethod
    def generate_id(client, identifier=None):
        """Generate an id for an argument."""
        host = "localhost"
        if client:
            host = client.remote.get("host") or host
        host = os.environ.get("RENKU_DOMAIN") or host

        if not identifier:
            identifier = str(uuid.uuid4())

        return urllib.parse.urljoin(
            "https://{host}".format(host=host),
            pathlib.posixpath.join("/runs", urllib.parse.quote(identifier, safe="")),
        )

    @classmethod
    def from_factory(cls, factory, client, commit, path):
        """Creates a ``Run`` from a ``CommandLineToolFactory``."""
        inputs = []
        arguments = []
        run_id = cls.generate_id(client)
        outputs = [_convert_cmd_output(o, factory, client, commit, run_id) for o in factory.outputs]

        if outputs:
            outputs, inputs_to_remove = zip(*outputs)
            outputs = list(outputs)

            for i in inputs_to_remove:
                # remove inputs that are actually outputs
                # note: a single input can represent multiple outputs
                # in case of repetition in the cli
                if not i:
                    continue
                if i in factory.inputs:
                    factory.inputs.remove(i)

        for i in factory.inputs:
            res = _convert_cmd_input(i, client, commit, run_id)

            if isinstance(res, CommandInput):
                inputs.append(res)
            else:
                arguments.append(res)

        return cls(
            id=run_id,
            client=client,
            commit=commit,
            path=path,
            command=" ".join(factory.baseCommand),
            successcodes=factory.successCodes,
            arguments=[_convert_cmd_binding(a, client, commit) for a in factory.arguments] + arguments,
            inputs=inputs,
            outputs=outputs,
            run_parameters=[_convert_run_parameter(a, run_id) for a in factory.run_parameters],
        )

    @property
    def activity(self):
        """Return the activity object."""
        return self._activity() if self._activity else None

    def to_argv(self):
        """Convert run into argv list."""
        argv = []

        if self.command:
            argv.extend(self.command.split(" "))

        arguments = self.inputs + self.outputs + self.arguments

        arguments = filter(lambda x: x.position, arguments)
        arguments = sorted(arguments, key=lambda x: x.position)
        argv.extend(e for a in arguments for e in a.to_argv())

        return argv

    def to_stream_repr(self):
        """Input/output stream representation."""
        stream_repr = []

        for input_ in self.inputs:
            if input_.mapped_to:
                stream_repr.append(input_.to_stream_repr())

        for output in self.outputs:
            if output.mapped_to:
                stream_repr.append(output.to_stream_repr())
        return stream_repr

    def update_id_and_label_from_commit_path(self, client, commit, path, is_subprocess=False):
        """Updates the _id and _label using supplied commit and path."""
        self.client = client
        if not self.commit:
            self.commit = commit
            if not is_subprocess:
                path = Path(os.path.abspath(path)).relative_to(self.client.path)
                self.path = path
                self._id = self.generate_id(client)
                self._label = self.default_label()

        if len(self.subprocesses) > 0:
            for s in self.subprocesses:
                s.process.update_id_and_label_from_commit_path(client, commit, path, is_subprocess=True)

    def add_subprocess(self, subprocess):
        """Adds a subprocess to this run."""
        process_order = 0
        if self.subprocesses:
            processes = [o.process for o in self.subprocesses]
            # Get position to insert based on dependencies
            process_order = bisect(processes, subprocess)
            if process_order < len(processes):
                # adjust ids of inputs inherited from latter subprocesses
                for i in range(len(processes), process_order, -1):
                    sp = self.subprocesses[i - 1]
                    sp._id = sp._id.replace(f"subprocess/{i}", f"subprocess/{i+1}")
                    sp.index += 1

                    for inp in self.inputs:
                        inp._id = inp._id.replace(f"/steps/step_{i}/", f"/steps/step_{i+1}/")
                    for outp in self.outputs:
                        outp._id = outp._id.replace(f"/steps/step_{i}/", f"/steps/step_{i+1}/")

        input_paths = [i.consumes.path for i in self.inputs]
        output_paths = [o.produces.path for o in self.outputs]

        for input_ in subprocess.inputs:
            if input_.consumes.path not in input_paths and input_.consumes.path not in output_paths:
                new_input = copy(input_)

                new_input._id = f"{self._id}/steps/step_{process_order + 1}/" f"{new_input.sanitized_id}"
                new_input.mapped_to = None

                matching_output = next((o for o in self.outputs if o.produces.path == new_input.consumes.path), None)

                if not matching_output:
                    self.inputs.append(new_input)
                    input_paths.append(new_input.consumes.path)

        for output in subprocess.outputs:
            if output.produces.path not in output_paths:
                new_output = copy(output)

                new_output._id = f"{self._id}/steps/step_{process_order + 1}/" f"{new_output.sanitized_id}"
                new_output.mapped_to = None
                self.outputs.append(new_output)
                output_paths.append(new_output.produces.path)

                matching_input = next((i for i in self.inputs if i.consumes.path == new_output.produces.path), None)
                if matching_input:
                    self.inputs.remove(matching_input)
                    input_paths.remove(matching_input.consumes.path)
        ordered_process = OrderedSubprocess(
            id=OrderedSubprocess.generate_id(self._id, process_order + 1), index=process_order + 1, process=subprocess
        )
        self.subprocesses.insert(process_order, ordered_process)

    def __lt__(self, other):
        """Compares two subprocesses order based on their dependencies."""
        a_inputs = set()
        b_outputs = set()

        for i in other.inputs:
            entity = i.consumes
            for subentity in entity.entities:
                a_inputs.add(subentity.path)

        for i in self.outputs:
            entity = i.produces
            for subentity in entity.entities:
                b_outputs.add(subentity.path)

        return a_inputs & b_outputs

    def __attrs_post_init__(self):
        """Calculate properties."""
        if self.client and not self._id:
            self._id = Run.generate_id(self.client)
        super().__attrs_post_init__()

        commit_not_set = not self.commit or self.commit.hexsha in self._label
        if commit_not_set and self.client and self.path and Path(self.path).exists():
            self.commit = self.client.find_previous_commit(self.path)

        # List order is not guaranteed when loading from JSON-LD
        self.subprocesses.sort()

    @classmethod
    def from_jsonld(cls, data):
        """Create an instance from JSON-LD data."""
        if isinstance(data, cls):
            return data
        if not isinstance(data, dict):
            raise ValueError(data)

        return RunSchema().load(data)

    def as_jsonld(self):
        """Create JSON-LD."""
        return RunSchema().dump(self)


@total_ordering
@attr.s(cmp=False,)
class OrderedSubprocess:
    """A subprocess with ordering."""

    _id = attr.ib(kw_only=True)

    index = attr.ib(kw_only=True, type=int)

    process = attr.ib(kw_only=True)

    @staticmethod
    def generate_id(parent_id, index):
        """Generate an id for an ``OrderedSubprocess``."""
        return f"{parent_id}/subprocess/{index}"

    def __lt__(self, other):
        """Compares two ordered subprocesses."""
        return self.index < other.index


class RunSchema(CommitMixinSchema):
    """Run schema."""

    class Meta:
        """Meta class."""

        rdf_type = [renku.Run, prov.Plan, prov.Entity]
        model = Run
        unknown = EXCLUDE

    command = fields.String(renku.command, missing=None)
    successcodes = fields.List(renku.successCodes, fields.Integer(), missing=[0])
    subprocesses = Nested(renku.hasSubprocess, nested="OrderedSubprocessSchema", missing=None, many=True)
    arguments = Nested(renku.hasArguments, CommandArgumentSchema, many=True, missing=None)
    inputs = Nested(renku.hasInputs, CommandInputSchema, many=True, missing=None)
    outputs = Nested(renku.hasOutputs, CommandOutputSchema, many=True, missing=None)
    run_parameters = Nested(renku.hasRunParameters, RunParameterSchema, many=True, missing=None)


class OrderedSubprocessSchema(JsonLDSchema):
    """OrderedSubprocess schema."""

    class Meta:
        """Meta class."""

        rdf_type = [renku.OrderedSubprocess]
        model = OrderedSubprocess
        unknown = EXCLUDE

    _id = fields.Id(init_name="id")
    index = fields.Integer(renku.index)
    process = Nested(renku.process, RunSchema)



# -*- coding: utf-8 -*-
#
# Copyright 2017-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Renku API Workflow Models.

Input and Output classes can be used to define inputs and outputs of a script
within the same script. Paths defined with these classes are added to explicit
inputs and outputs in the workflow's metadata. For example, the following
mark a ``data/data.csv`` as an input to the script:

.. code-block:: python

    from renku.api import Input

    with open(Input("data/data.csv")) as input_data:
        for line in input_data:
            print(line)


Users can track parameters' values in a workflow by defining them using
``Parameter`` function.

.. code-block:: python

    from renku.api import Parameter

    nc = Parameter(name="n_components", value=10)

"""

from os import PathLike
from pathlib import Path

from renku.api.models.project import ensure_project_context
from renku.core.models.cwl.command_line_tool import (
    add_indirect_parameter,
    get_indirect_inputs_path,
    get_indirect_outputs_path,
)


class _PathBase(PathLike):
    @ensure_project_context
    def __init__(self, path, project):
        self._path = Path(path)

        indirect_list_path = self._get_indirect_list_path(project.path)
        indirect_list_path.parent.mkdir(exist_ok=True, parents=True)
        with open(indirect_list_path, "a") as file_:
            file_.write(str(path))
            file_.write("\n")

    @staticmethod
    def _get_indirect_list_path(project_path):
        raise NotImplementedError

    def __fspath__(self):
        """Abstract method of PathLike."""
        return str(self._path)

    @property
    def path(self):
        """Return path of file."""
        return self._path


class Input(_PathBase):
    """API Input model."""

    @staticmethod
    def _get_indirect_list_path(project_path):
        return get_indirect_inputs_path(project_path)


class Output(_PathBase):
    """API Output model."""

    @staticmethod
    def _get_indirect_list_path(project_path):
        return get_indirect_outputs_path(project_path)


@ensure_project_context
def parameter(name, value, project):
    """Store parameter's name and value."""
    add_indirect_parameter(project.path, name=name, value=value)

    return value


Parameter = parameter



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility for working with HTTP session."""
from contextlib import contextmanager

import requests
from requests.adapters import HTTPAdapter
from requests.exceptions import RequestException
from urllib3.util.retry import Retry

from renku.core.errors import RenkuException


@contextmanager
def retry(total_requests=10, backoff_factor=1, statuses=(500, 502, 503, 504, 429)):
    """Default HTTP session for requests."""
    _session = requests.Session()

    retries = Retry(total=total_requests, backoff_factor=backoff_factor, status_forcelist=list(statuses))

    _session.mount("http://", HTTPAdapter(max_retries=retries))
    _session.mount("https://", HTTPAdapter(max_retries=retries))

    try:
        yield _session
    except RequestException as e:
        raise RenkuException("renku operation failed due to network connection failure") from e



# -*- coding: utf-8 -*-
#
# Copyright 2017-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Migrate project to the latest Renku version."""

from renku.core.management.migrate import (
    is_docker_update_possible,
    is_migration_required,
    is_project_unsupported,
    is_renku_project,
    is_template_update_possible,
    migrate,
)

from .client import pass_local_client

SUPPORTED_RENKU_PROJECT = 0
MIGRATION_REQUIRED = 1
UNSUPPORTED_PROJECT = 2
NON_RENKU_REPOSITORY = 3


@pass_local_client
def migrations_check(client):
    """Public function for a migrations check."""

    template_update_possible, current_version, new_version = is_template_update_possible(client)
    return (
        is_migration_required(client),
        not is_project_unsupported(client),
        template_update_possible,
        current_version,
        new_version,
        bool(client.project.automated_update),
        is_docker_update_possible(client),
    )


@pass_local_client
def migrations_versions(client):
    """Return source and destination migration versions."""
    from renku import __version__

    return __version__, client.latest_agent


@pass_local_client(clean=True, commit=True, commit_empty=False)
def migrate_project(
    client,
    force_template_update=False,
    skip_template_update=False,
    skip_docker_update=False,
    skip_migrations=False,
    progress_callback=None,
    commit_message=None,
):
    """Migrate all project's entities."""
    return migrate(
        client=client,
        force_template_update=force_template_update,
        skip_template_update=skip_template_update,
        skip_docker_update=skip_docker_update,
        skip_migrations=skip_migrations,
        progress_callback=progress_callback,
    )


@pass_local_client(clean=True, commit=False)
def migrate_project_no_commit(
    client,
    force_template_update=False,
    skip_template_update=False,
    skip_docker_update=False,
    skip_migrations=False,
    progress_callback=None,
):
    """Migrate all project's entities but do not commit changes."""
    return migrate(
        client=client,
        force_template_update=force_template_update,
        skip_template_update=skip_template_update,
        skip_docker_update=skip_docker_update,
        skip_migrations=skip_migrations,
        progress_callback=progress_callback,
    )


@pass_local_client
def check_project(client):
    """Check if repository is a renku project, unsupported, or requires migration."""
    if not is_renku_project(client):
        return NON_RENKU_REPOSITORY
    elif is_migration_required(client):
        return MIGRATION_REQUIRED
    elif is_project_unsupported(client):
        return UNSUPPORTED_PROJECT

    return SUPPORTED_RENKU_PROJECT



# -*- coding: utf-8 -*-
#
# Copyright 2019-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Project initialization tests."""

import shutil
from pathlib import Path
from tempfile import TemporaryDirectory

import pkg_resources
import pytest

from renku.core import errors
from renku.core.commands.init import (
    TEMPLATE_MANIFEST,
    create_from_template,
    fetch_template_from_git,
    read_template_manifest,
    validate_template,
)
from renku.core.management.config import RENKU_HOME
from renku.core.management.migrate import migrate
from tests.utils import raises

template_local = Path(pkg_resources.resource_filename("renku", "templates"))


@pytest.mark.integration
def test_fetch_template_from_git(template):
    """Test fetching a template.

    It fetches a template from remote and verifies that the manifest
    file is there.
    """
    # ? Can't parametrize with fixtures yet
    # ? REF: https://docs.pytest.org/en/stable/proposals/parametrize_with_fixtures.html
    parametrize = [
        {"url": template["url"], "ref": template["ref"], "error": None},
        {"url": "fake", "ref": template["ref"], "error": errors.GitError},
        {"url": template["url"], "ref": "fake", "error": errors.GitError},
    ]
    for parameters in parametrize:
        with TemporaryDirectory() as tempdir:
            with raises(parameters["error"]):
                manifest_file, _ = fetch_template_from_git(parameters["url"], parameters["ref"], Path(tempdir))
                assert Path(tempdir) / TEMPLATE_MANIFEST == manifest_file
                assert manifest_file.exists()


def test_read_template_manifest():
    """Test reading template manifest file.

    It creates a fake manifest file and it verifies it's read property.
    """
    with TemporaryDirectory() as tempdir:
        template_file = Path(tempdir) / TEMPLATE_MANIFEST

        # error on missing template file
        with raises(errors.InvalidTemplateError):
            manifest = read_template_manifest(Path(tempdir), checkout=False)

        template_file.touch(exist_ok=True)
        # error on invalid template file
        with raises(errors.InvalidTemplateError):
            manifest = read_template_manifest(Path(tempdir), checkout=False)

        template_file.write_text(
            "-\n"
            "  folder: first\n"
            "  name: Basic Project 1\n"
            "  description: Description 1\n"
            "  variables: {}\n"
            "-\n"
            "  folder: second\n"
            "  name: Basic Project 2\n"
            "  description: Description 2\n"
            "  variables:\n"
            "    custom: Custom Value\n"
        )

        manifest = read_template_manifest(Path(tempdir), checkout=False)
        assert 2 == len(manifest)
        assert "first" == manifest[0]["folder"]
        assert "second" == manifest[1]["folder"]
        assert "Basic Project 1" == manifest[0]["name"]
        assert "Description 2" == manifest[1]["description"]

        variables1 = manifest[0]["variables"]
        variables2 = manifest[1]["variables"]
        assert 0 == len(variables1)
        assert 1 == len(variables2)
        assert 1 == len(variables2.keys())
        assert "custom" in variables2.keys()
        assert "Custom Value" == variables2["custom"]

        template_file.write_text("-\n" "  folder: first\n" "  description: Description 1\n")
        with raises(errors.InvalidTemplateError):
            manifest = read_template_manifest(Path(tempdir), checkout=False)

        template_file.write_text("-\n" "  name: Basic Project 2\n" "  description: Description 2\n")
        with raises(errors.InvalidTemplateError):
            manifest = read_template_manifest(Path(tempdir), checkout=False)


@pytest.mark.integration
def test_fetch_template_from_git_and_read_manifest(template):
    """Test template fetch and manifest reading.

    It fetches a local template, reads the manifest, checkouts the
    template folders and verify they exist.
    """
    with TemporaryDirectory() as tempdir:
        template_path = Path(tempdir)
        fetch_template_from_git(template["url"], template["ref"], template_path)
        manifest = read_template_manifest(template_path, checkout=True)
        for template in manifest:
            template_folder = template_path / template["folder"]
            assert template_folder.exists()


def test_validate_template():
    """Test template validation.

    It validates each local template.
    """
    with TemporaryDirectory() as tempdir:
        temppath = Path(tempdir)
        # file error
        with raises(errors.InvalidTemplateError):
            validate_template(temppath)

        # folder error
        shutil.rmtree(str(tempdir))
        renku_dir = temppath / RENKU_HOME
        renku_dir.mkdir(parents=True)
        with raises(errors.InvalidTemplateError):
            validate_template(temppath)

        # valid template
        shutil.rmtree(str(tempdir))
        shutil.copytree(str(template_local), str(tempdir))
        manifest = read_template_manifest(Path(tempdir))
        for template in manifest:
            template_folder = temppath / template["folder"]
            assert validate_template(template_folder) is True


def test_create_from_template(local_client, template):
    """Test repository creation from a template.

    It creates a renku projects from one of the local templates and it verifies
    the data are properly copied to the new renku project folder.
    """
    with TemporaryDirectory() as tempdir:
        temppath = Path(tempdir) / "local"
        shutil.copytree(str(template_local), str(temppath))
        manifest = read_template_manifest(temppath)
        template_path = temppath / manifest[0]["folder"]
        create_from_template(
            template_path, local_client, "name", {**template["default_metadata"], **template["metadata"]}
        )
        template_files = [
            f
            for f in local_client.path.glob("**/*")
            if ".git" not in str(f)
            and not str(f).endswith(".renku/metadata.yml")
            and not str(f).endswith(".renku/template_checksums.json")
        ]
        for template_file in template_files:
            expected_file = template_path / template_file.relative_to(local_client.path)
            assert expected_file.exists()


def test_template_filename(local_client, template):
    """Test using a template with dynamic filenames.
    """
    with TemporaryDirectory() as tempdir:
        template_folder = Path(tempdir) / "first"

        template_folder.mkdir(parents=True)

        template_file = template_folder / "{{ name }}.r"
        template_file.write_text("{{ name }}")

        (local_client.path / ".renku").mkdir()

        create_from_template(template_folder, local_client, name="test", metadata=template["default_metadata"])

        assert (local_client.path / "test.r").exists()


def test_update_from_template(local_client, template_update):
    """Test repository update from a template."""
    res = template_update()
    project_files = res["project_files"]
    template_files = res["template_files"]

    project_files_before = {p: p.read_text() for p in project_files if not p.is_dir()}

    for p in template_files:
        if p.is_dir():
            continue
        p.write_text(f"{p.read_text()}\nmodified")

    migrate(local_client, skip_docker_update=True)

    for p in project_files:
        if p.is_dir():
            continue
        content = project_files_before[p]
        new_content = p.read_text()
        assert content != new_content


def test_update_from_template_with_modified_files(local_client, template_update):
    """Test repository update from a template with modified local files."""
    res = template_update()
    project_files = res["project_files"]
    template_files = res["template_files"]

    project_files_before = {p: p.read_text() for p in project_files if not p.is_dir()}

    for p in template_files:
        if p.is_dir():
            continue
        p.write_text(f"{p.read_text()}\nmodified")

    # NOTE: modify local file
    modified_file = next(f for f in project_files if str(f).endswith("README.md"))
    modified_local_content = modified_file.read_text() + "\nlocal modification"
    modified_file.write_text(modified_local_content)

    # NOTE: delete local file
    deleted_file = next(f for f in project_files if str(f).endswith("README.md"))
    deleted_file.unlink()

    migrate(local_client, skip_docker_update=True)

    for p in project_files:
        if p.is_dir():
            continue
        if p == deleted_file:
            assert not p.exists()
            continue

        content = project_files_before[p]
        new_content = p.read_text()

        if p == modified_file:
            assert modified_local_content == new_content
        else:
            assert content != new_content


def test_update_from_template_with_immutable_modified_files(local_client, mocker, template_update):
    """Test repository update from a template with modified local immutable files."""
    res = template_update(immutable_files=["README.md"])
    project_files = res["project_files"]
    template_files = res["template_files"]

    for p in template_files:
        if p.is_dir():
            continue
        p.write_text(f"{p.read_text()}\nmodified")

    # NOTE: modify local file
    modified_file = next(f for f in project_files if str(f).endswith("README.md"))
    modified_local_content = modified_file.read_text() + "\nlocal modification"
    modified_file.write_text(modified_local_content)

    with pytest.raises(
        errors.TemplateUpdateError, match=r"Can't update template as immutable template file .* has local changes."
    ):
        migrate(local_client)


def test_update_from_template_with_immutable_deleted_files(local_client, mocker, template_update):
    """Test repository update from a template with deleted local immutable files."""
    res = template_update(immutable_files=["README.md"])
    project_files = res["project_files"]
    template_files = res["template_files"]

    for p in template_files:
        if p.is_dir():
            continue
        p.write_text(f"{p.read_text()}\nmodified")

    # NOTE: modify local file
    deleted_file = next(f for f in project_files if str(f).endswith("README.md"))
    deleted_file.unlink()

    with pytest.raises(
        errors.TemplateUpdateError, match=r"Can't update template as immutable template file .* has local changes."
    ):
        migrate(local_client)


def test_update_template_dockerfile(local_client, mocker, template_update):
    """Test repository Dockerfile update."""
    template_update(docker=True, after_template_version="0.0.1")

    import renku

    mocker.patch.object(renku, "__version__", "0.0.2")

    migrate(local_client)

    dockerfile = (local_client.path / "Dockerfile").read_text()
    assert "0.0.2" in dockerfile


def test_update_from_template_with_new_variable(local_client, mocker, template_update):
    """Test repository update from a template with a new template variable required."""
    res = template_update()
    manifest = res["manifest"]
    manifest_path = res["manifest_path"]
    template_files = res["template_files"]

    # NOTE: Add new template variable
    manifest[0]["variables"]["__new_arbitrary_template_value__"] = None
    fetch_template = mocker.patch("renku.core.commands.init.fetch_template")
    fetch_template.return_value = (manifest, manifest_path, "renku", "0.0.2")

    for p in template_files:
        if p.is_dir():
            continue
        p.write_text(f"{p.read_text()}\nmodified")

    with pytest.raises(errors.TemplateUpdateError, match=r".*Can't update template, it now requires variable.*"):
        migrate(local_client)



# -*- coding: utf-8 -*-
#
# Copyright 2018-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test projects API."""
from datetime import timezone

from freezegun import freeze_time

from renku.core.models.projects import Project


def test_project_serialization(client):
    """Test project serialization with JSON-LD context."""
    from renku.core.management.migrate import SUPPORTED_PROJECT_VERSION

    with freeze_time("2017-03-01T08:00:00.000000+00:00") as frozen_time:
        project_time = frozen_time().replace(tzinfo=timezone.utc)
        project = Project(name="demo", client=client)
        assert project.name == "demo"
        assert project.created == project_time

    data = project.as_jsonld()
    assert "http://schema.org/Project" in data["@type"]
    assert "http://www.w3.org/ns/prov#Location" in data["@type"]

    assert "demo" == data["http://schema.org/name"]
    assert project_time.isoformat("T") == data["http://schema.org/dateCreated"]
    assert str(SUPPORTED_PROJECT_VERSION) == data["http://schema.org/schemaVersion"]


def test_project_creator_deserialization(client, project):
    """Check that the correct creator is returned on deserialization."""
    from renku.core.models.provenance.agents import Person

    # modify the project metadata to change the creator
    project = client.project
    project.creator = Person(email="johndoe@example.com", name="Johnny Doe")
    project.to_yaml()
    client.repo.git.commit("-a", "--amend", "-C", "HEAD", "--author", "Johnny Doe <johndoe@example.com>", "--no-verify")

    # the project creator should always be the one in the metadata
    assert "johndoe@example.com" == client.project.creator.email
    assert "Johnny Doe" == client.project.creator.name
    assert client.project.creator.label == client.project.creator.name

    # Remove the creator from metadata
    project = client.project
    project.creator = None
    project.to_yaml()
    client.repo.git.commit("-a", "--amend", "-C", "HEAD", "--author", "Jane Doe <janedoe@example.com>", "--no-verify")

    # now the creator should be the one from the commit
    project = Project.from_yaml(client.renku_metadata_path, client=client)
    assert "janedoe@example.com" == project.creator.email
    assert "Jane Doe" == project.creator.name
    assert project.creator.label == project.creator.name



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Define custom JSON encoder."""

import datetime
import functools
import json
from uuid import UUID


class JSONEncoder(json.JSONEncoder):
    """Support for custom data formats."""

    def default(self, obj):
        """Encode more types."""
        if isinstance(obj, UUID):
            return obj.hex
        elif isinstance(obj, datetime.datetime):
            return obj.isoformat()

        return super().default(obj)


dumps = functools.partial(json.dumps, cls=JSONEncoder)



# -*- coding: utf-8 -*-
#
# Copyright 2017-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku doctor tests."""
from pathlib import Path

from renku.cli import cli


def test_new_project_is_ok(runner, project):
    """Test renku doctor initially is OK on a new project."""
    # Initially, every thing is OK
    result = runner.invoke(cli, ["doctor"])
    assert 0 == result.exit_code
    assert "Everything seems to be ok." in result.output


def test_git_hooks_not_available(runner, project):
    """Test detection of not-installed git hooks."""
    result = runner.invoke(cli, ["githooks", "uninstall"])
    assert 0 == result.exit_code

    result = runner.invoke(cli, ["doctor"])
    assert 1 == result.exit_code
    assert "Git hooks are not installed." in result.output


def test_git_hooks_modified(runner, project):
    """Test detection of modified git hooks."""
    result = runner.invoke(cli, ["githooks", "install", "--force"])
    assert 0 == result.exit_code

    hook_path = Path(project) / ".git" / "hooks" / "pre-commit"
    lines = hook_path.read_text().split("\n")

    # Append some more commands
    appended = lines + ["# Some more commands", "ls"]
    hook_path.write_text("\n".join(appended))

    # Check passes as long as Renku hook is not modified
    result = runner.invoke(cli, ["doctor"])
    assert 0 == result.exit_code
    assert "Everything seems to be ok." in result.output

    # Modify Renku hook
    modified = [line for line in lines if "# END RENKU HOOK." not in line]
    hook_path.write_text("\n".join(modified))

    result = runner.invoke(cli, ["doctor"])
    assert 1 == result.exit_code
    assert "Git hooks are outdated or not installed." in result.output


def test_lfs_broken_history(runner, client, tmp_path):
    """Test lfs migrate info check on a broken history."""
    big_file = tmp_path / "big-file.bin"
    with open(big_file, "w") as file_:
        file_.seek(client.minimum_lfs_file_size)
        file_.write("some-data")

    # Add a file without adding it to LFS
    result = runner.invoke(
        cli,
        ["--no-external-storage", "dataset", "add", "--create", "new-dataset", str(big_file)],
        catch_exceptions=False,
    )
    assert 0 == result.exit_code

    result = runner.invoke(cli, ["doctor"])
    assert 1 == result.exit_code
    assert "Git history contains large files" in result.output
    assert "*.bin" in result.output

    # Exclude *.ipynb files from LFS in .renkulfsignore
    (client.path / client.RENKU_LFS_IGNORE_PATH).write_text("\n".join(["*swp", "*.bin", ".DS_Store"]))

    result = runner.invoke(cli, ["doctor"])
    assert 0 == result.exit_code
    assert "Git history contains large files" not in result.output



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service view decorators."""
import re
from functools import wraps

from flask import jsonify, request
from flask_apispec import doc
from git import GitCommandError, GitError
from marshmallow import ValidationError
from redis import RedisError
from sentry_sdk import capture_exception
from werkzeug.exceptions import HTTPException

from renku.core.errors import (
    DockerfileUpdateError,
    MigrationError,
    MigrationRequired,
    RenkuException,
    TemplateUpdateError,
    UninitializedProject,
)
from renku.service.cache import cache
from renku.service.config import (
    GIT_ACCESS_DENIED_ERROR_CODE,
    GIT_UNKNOWN_ERROR_CODE,
    INTERNAL_FAILURE_ERROR_CODE,
    INVALID_HEADERS_ERROR_CODE,
    INVALID_PARAMS_ERROR_CODE,
    REDIS_EXCEPTION_ERROR_CODE,
    RENKU_EXCEPTION_ERROR_CODE,
)
from renku.service.serializers.headers import UserIdentityHeaders
from renku.service.utils.squash import squash
from renku.service.views import error_response


def requires_identity(f):
    """Wrapper which indicates that route requires user identification."""
    # noqa
    @wraps(f)
    def decorated_function(*args, **kws):
        """Represents decorated function."""
        try:
            user = UserIdentityHeaders().load(request.headers)
        except (ValidationError, KeyError) as e:
            capture_exception(e)

            err_message = "user identification is incorrect or missing"
            return jsonify(error={"code": INVALID_HEADERS_ERROR_CODE, "reason": err_message})

        return f(user, *args, **kws)

    return decorated_function


def handle_redis_except(f):
    """Wrapper which handles Redis exceptions."""
    # noqa
    @wraps(f)
    def decorated_function(*args, **kwargs):
        """Represents decorated function."""
        try:
            return f(*args, **kwargs)
        except (RedisError, OSError) as e:
            capture_exception(e)

            return jsonify(error={"code": REDIS_EXCEPTION_ERROR_CODE, "reason": e.messages})

    return decorated_function


@handle_redis_except
def requires_cache(f):
    """Wrapper which injects cache object into view."""
    # noqa
    @wraps(f)
    def decorated_function(*args, **kwargs):
        """Represents decorated function."""
        return f(cache, *args, **kwargs)

    return decorated_function


def handle_schema_except(f):
    """Wrapper which handles Schema.load exceptions."""
    # noqa
    @wraps(f)
    def decorated_function(*args, **kwargs):
        """Represents decorated function."""
        try:
            return f(*args, **kwargs)
        except KeyError as e:
            capture_exception(e)

            if e.args and len(e.args) > 0:
                return jsonify(error={"code": INVALID_PARAMS_ERROR_CODE, "reason": f'missing parameter "{e.args[0]}"'})

            raise

    return decorated_function


def handle_validation_except(f):
    """Wrapper which handles marshmallow `ValidationError`."""

    # noqa
    @wraps(f)
    def decorated_function(*args, **kwargs):
        """Represents decorated function."""
        try:
            return f(*args, **kwargs)
        except ValidationError as e:
            capture_exception(e)

            reasons = [f"`{key}` - {', '.join(value)}" for key, value in squash(e.messages).items()]
            error_message = f"Validation error: {'; '.join(reasons)}"

            return error_response(INVALID_PARAMS_ERROR_CODE, error_message)

    return decorated_function


def handle_renku_except(f):
    """Wrapper which handles `RenkuException`."""
    # noqa
    @wraps(f)
    def decorated_function(*args, **kwargs):
        """Represents decorated function."""
        try:
            return f(*args, **kwargs)
        except RenkuException as e:
            capture_exception(e)

            err_response = {
                "code": RENKU_EXCEPTION_ERROR_CODE,
                "reason": str(e),
            }

            if isinstance(e, UninitializedProject):
                err_response["project_initialization_required"] = True

            if isinstance(e, MigrationRequired):
                err_response["migration_required"] = True

            return jsonify(error=err_response)

    return decorated_function


def handle_migration_except(f):
    """Wrapper which handles exceptions during migrations."""
    # noqa
    @wraps(f)
    def decorated_function(*args, **kwargs):
        """Represents decorated function."""
        try:
            return f(*args, **kwargs)
        except (TemplateUpdateError, DockerfileUpdateError, MigrationError) as e:
            err_response = {
                "code": RENKU_EXCEPTION_ERROR_CODE,
                "reason": str(e),
                "template_update_failed": isinstance(e, TemplateUpdateError),
                "dockerfile_update_failed": isinstance(e, DockerfileUpdateError),
                "migrations_failed": isinstance(e, MigrationError),
            }

            return jsonify(error=err_response)

    return decorated_function


def handle_git_except(f):
    """Wrapper which handles `RenkuException`."""
    # noqa
    @wraps(f)
    def decorated_function(*args, **kwargs):
        """Represents decorated function."""
        try:
            return f(*args, **kwargs)
        except GitCommandError as e:
            capture_exception(e)

            error_code = GIT_ACCESS_DENIED_ERROR_CODE if "Access denied" in e.stderr else GIT_UNKNOWN_ERROR_CODE

            if "is this a git repository?" in e.stderr:
                error_reason = "Repository could not be found"
            elif "Access denied" in e.stderr:
                error_reason = "Repository could not be accessed - Do you have access rights?"
            else:
                error_reason = format(" ".join(e.stderr.strip().split("\n")))

                # strip oauth tokens
                error_reason_safe = re.sub("^(.+oauth2:)[^@]+(@.+)$", r"\1<token-hidden>\2", error_reason)
                error_reason = f"git error: {error_reason_safe}"

            return error_response(error_code, error_reason)

    return decorated_function


def accepts_json(f):
    """Wrapper which ensures only JSON payload can be in request."""
    # noqa
    @wraps(f)
    def decorated_function(*args, **kwargs):
        """Represents decorated function."""
        if "Content-Type" not in request.headers:
            return jsonify(error={"code": INVALID_HEADERS_ERROR_CODE, "reason": "invalid request headers"})

        header_check = request.headers["Content-Type"] == "application/json"

        if not request.is_json or not header_check:
            return jsonify(error={"code": INVALID_HEADERS_ERROR_CODE, "reason": "invalid request payload"})

        return f(*args, **kwargs)

    return decorated_function


def handle_base_except(f):
    """Wrapper which handles base exceptions."""
    # noqa
    @wraps(f)
    def decorated_function(*args, **kwargs):
        """Represents decorated function."""
        try:
            return f(*args, **kwargs)

        except HTTPException as e:  # handle general werkzeug exception
            capture_exception(e)

            error_message = f"Failed to contact external service. Received response {e.code} ({e.description})."
            return error_response(INTERNAL_FAILURE_ERROR_CODE, error_message)

        except GitError as e:
            capture_exception(e)

            error_message = "Failed to execute git operation."
            return error_response(INTERNAL_FAILURE_ERROR_CODE, error_message)

        except (Exception, BaseException, OSError, IOError) as e:
            capture_exception(e)

            internal_error = "internal error"
            if hasattr(e, "stderr"):
                internal_error += ": {0}".format(" ".join(e.stderr.strip().split("\n")))

            return error_response(INTERNAL_FAILURE_ERROR_CODE, internal_error)

    return decorated_function


def header_doc(description, tags=()):
    """Wrap additional OpenAPI header description for an endpoint."""
    return doc(
        description=description,
        params={
            "Authorization": {
                "description": (
                    "Used for users git oauth2 access. " "For example: " "```Authorization: Bearer asdf-qwer-zxcv```"
                ),
                "in": "header",
                "type": "string",
                "required": True,
            },
            "Renku-User-Id": {
                "description": (
                    "Used for identification of the users. "
                    "For example: "
                    "```Renku-User-Id: sasdsa-sadsd-gsdsdh-gfdgdsd```"
                ),
                "in": "header",
                "type": "string",
                "required": True,
            },
            "Renku-User-FullName": {
                "description": (
                    "Used for commit author signature. " "For example: " "```Renku-User-FullName: Rok Roskar```"
                ),
                "in": "header",
                "type": "string",
                "required": True,
            },
            "Renku-User-Email": {
                "description": (
                    "Used for commit author signature. " "For example: " "```Renku-User-Email: dev@renkulab.io```"
                ),
                "in": "header",
                "type": "string",
                "required": True,
            },
        },
        tags=list(tags),
    )


def handle_common_except(f):
    """Handle common exceptions."""
    # noqa
    @wraps(f)
    def dec(*args, **kwargs):
        """Decorated function."""

        @handle_base_except
        @handle_validation_except
        @handle_renku_except
        @handle_git_except
        def _wrapped(*args_, **kwargs_):
            return f(*args_, **kwargs_)

        return _wrapped(*args, **kwargs)

    return dec



# -*- coding: utf-8 -*-
#
# Copyright 2018-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Represent parameters from the Common Workflow Language."""

import attr


def convert_default(value):
    """Convert a default value."""
    return value


@attr.s
class _IdMixin(object):
    """Define id field."""

    id = attr.ib(default=None)


@attr.s
class Parameter(object):
    """Define an input or output parameter to a process."""

    streamable = attr.ib(default=None, converter=bool)


@attr.s
class InputParameter(_IdMixin, Parameter):
    """An input parameter."""

    type = attr.ib(default="string")
    description = attr.ib(default=None)
    default = attr.ib(default=None, converter=convert_default)
    inputBinding = attr.ib(default=None)


@attr.s
class CommandLineBinding(object):
    """Define the binding behavior when building the command line."""

    position = attr.ib(default=None)  # int
    prefix = attr.ib(default=None)  # int
    separate = attr.ib(default=True, type=bool)
    itemSeparator = attr.ib(default=None)  # str
    valueFrom = attr.ib(default=None)  # str | Expression
    shellQuote = attr.ib(default=True, type=bool)

    def to_argv(self, default=None):
        """Format command line binding as shell argument."""
        if self.valueFrom is not None:
            if self.valueFrom.startswith("$("):
                raise NotImplementedError()
            value = self.valueFrom
        else:
            value = default

        def _convert(value):
            """Convert value to a argument list."""
            if self.prefix:
                if self.separate:
                    return [self.prefix, str(value)]
                else:
                    return [self.prefix + str(value)]
            else:
                return [str(value)]

        if self.prefix is None and not self.separate:
            raise ValueError("Can not separate an empty prefix.")

        if isinstance(value, list):
            if self.itemSeparator and value:
                value = self.itemSeparator.join([str(v) for v in value])
            elif value:
                return [a for v in value for a in _convert(v)]
        elif (value is True or value is None) and self.prefix:
            return [self.prefix]
        elif value is False or value is None or (value is True and not self.prefix):
            return []

        return _convert(value)


@attr.s
class CommandInputParameter(InputParameter):
    """An input parameter for a CommandLineTool."""

    inputBinding = attr.ib(
        default=None,
        converter=lambda data: CommandLineBinding(**data)
        if not isinstance(data, CommandLineBinding) and data is not None
        else data,
    )

    @classmethod
    def from_cwl(cls, data):
        """Create instance from type definition."""
        if not isinstance(data, dict):
            data = {"type": data}
        return cls(**data)

    def to_argv(self, **kwargs):
        """Format command input parameter as shell argument."""
        return self.inputBinding.to_argv(default=self.default, **kwargs) if self.inputBinding else []


@attr.s
class OutputParameter(_IdMixin, Parameter):
    """An output parameter."""

    type = attr.ib(default="string")
    description = attr.ib(default=None)
    format = attr.ib(default=None)
    outputBinding = attr.ib(default=None)


@attr.s
class CommandOutputBinding(object):
    """Define the binding behavior for outputs."""

    glob = attr.ib(default=None)  # null, string, Expression, array[string]
    # loadContents, outputEval


@attr.s
class CommandOutputParameter(OutputParameter):
    """Define an output parameter for a CommandLineTool."""

    outputBinding = attr.ib(
        default=None,
        converter=lambda data: CommandOutputBinding(**data)
        if not isinstance(data, CommandOutputBinding) and data is not None
        else data,
    )


@attr.s
class WorkflowOutputParameter(OutputParameter):
    """Define an output parameter for a Workflow."""

    outputSource = attr.ib(default=None)


@attr.s
class RunParameter:
    """Define a parameter for a Workflow that is not passed via command-line."""

    name = attr.ib(default=None)
    value = attr.ib(default=None)



# -*- coding: utf-8 -*-
#
# Copyright 2019-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service rpc serializers."""
from renku.service.serializers.datasets import DatasetListResponseRPC
from renku.service.views import error_response, result_response


def test_error_response(svc_client):
    """Test error response utility."""
    err_code = 0
    err_reason = "test error"
    response = error_response(err_code, err_reason).json

    assert response
    assert {"error"} == set(response.keys())
    assert {"code", "reason"} == set(response["error"].keys())

    assert err_code == response["error"]["code"]
    assert err_reason == response["error"]["reason"]


def test_result_response(svc_client):
    """Test result response utility."""
    ctx = {"datasets": [{"name": "my-dataset"}]}
    response = result_response(DatasetListResponseRPC(), ctx).json

    assert response
    assert {"result"} == set(response.keys())
    assert {"datasets"} == set(response["result"].keys())

    expected = ctx["datasets"][0]
    received = response["result"]["datasets"][0]

    assert expected["name"] == received["name"]


def test_result_response_with_none(svc_client):
    """Test result response with None value."""
    response = result_response(DatasetListResponseRPC(), None).json

    assert response
    assert {"result"} == set(response.keys())
    assert response["result"] is None


def test_result_response_with_empty_dict(svc_client):
    """Test result response with empty value."""
    response = result_response(DatasetListResponseRPC(), {}).json

    assert response
    assert {"result"} == set(response.keys())
    assert {} == response["result"]


def test_result_response_with_empty_tuple(svc_client):
    """Test result response with empty value."""
    response = result_response(DatasetListResponseRPC(), ()).json

    assert response
    assert {"result"} == set(response.keys())
    assert {} == response["result"]



# -*- coding: utf-8 -*-
#
# Copyright 2018-2020- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Manage the set of CWL files created by ``renku`` commands.

Manipulating workflows
~~~~~~~~~~~~~~~~~~~~~~

Listing workflows:

.. code-block:: console

    $ renku workflow ls
    26be2e8d66f74130a087642768f2cef0_rerun.yaml:
    199c4b9d462f4b27a4513e5e55f76eb2_cat.yaml:
    9bea2eccf9624de387d9b06e61eec0b6_rerun.yaml:
    b681b4e229764ceda161f6551370af12_update.yaml:
    25d0805243e3468d92a3786df782a2c4_rerun.yaml:

Each ``*.yaml`` file corresponds to a renku run/update/rerun execution.

Exporting workflows:

You can export the workflow to create a file as Common Workflow Language
by using:

.. code-block:: console

    $ renku workflow set-name create output_file
    baseCommand:
    - cat
    class: CommandLineTool
    cwlVersion: v1.0
    id: 22943eca-fa4c-4f3b-a92d-f6ac7badc0d2
    inputs:
    - default:
        class: File
        path: /home/user/project/intermediate
    id: inputs_1
    inputBinding:
        position: 1
    type: File
    - default:
        class: File
        path: /home/user/project/intermediate2
    id: inputs_2
    inputBinding:
        position: 2
    type: File
    outputs:
    - id: output_stdout
    streamable: false
    type: stdout
    requirements:
    InitialWorkDirRequirement:
        listing:
        - entry: $(inputs.inputs_1)
        entryname: intermediate
        writable: false
        - entry: $(inputs.inputs_2)
        entryname: intermediate2
        writable: false
    stdout: output_file

You can use ``--revision`` to specify the revision of the output file to
generate the workflow for. You can also export to a file directly with
``-o <path>``.
"""

from collections import defaultdict
from pathlib import Path

import click

from renku.core.commands.client import pass_local_client
from renku.core.commands.graph import Graph
from renku.core.models.workflow.converters.cwl import CWLConverter

# TODO: Finish refactoring (ticket #703)


def _ref(name):
    """Return workflow reference name."""
    return "workflows/{0}".format(name)


def _deref(ref):
    """Remove workflows prefix."""
    assert ref.startswith("workflows/")
    return ref[len("workflows/") :]


@click.group()
def workflow():
    """Workflow commands."""
    pass


@workflow.command("ls")
@pass_local_client(requires_migration=True)
def list_workflows(client):
    """List or manage workflows with subcommands."""
    from renku.core.models.refs import LinkReference

    names = defaultdict(list)
    for ref in LinkReference.iter_items(client, common_path="workflows"):
        names[ref.reference.name].append(ref.name)

    for path in client.workflow_path.glob("*.yaml"):
        click.echo(
            "{path}: {names}".format(
                path=path.name, names=", ".join(click.style(_deref(name), fg="green") for name in names[path.name]),
            )
        )


def validate_path(ctx, param, value):
    """Detect a workflow path if it is not passed."""
    client = ctx.obj

    if value is None:
        from renku.core.models.provenance.activities import ProcessRun

        activity = client.process_commit()

        if not isinstance(activity, ProcessRun):
            raise click.BadParameter("No tool was found.")

        return activity.path

    return value


@workflow.command("set-name")
@click.argument("name", metavar="<name>")
@click.argument(
    "path",
    metavar="<path>",
    type=click.Path(exists=True, dir_okay=False),
    callback=validate_path,
    default=None,
    required=False,
)
@click.option("--force", is_flag=True, help="Override the existence check.")
@pass_local_client(clean=True, commit=True)
def set_name(client, name, path, force):
    """Sets the <name> for remote <path>."""
    from renku.core.models.refs import LinkReference

    LinkReference.create(client=client, name=_ref(name), force=force).set_reference(path)


@workflow.command()
@click.argument("old", metavar="<old>")
@click.argument("new", metavar="<new>")
@click.option("--force", is_flag=True, help="Override the existence check.")
@pass_local_client(clean=True, commit=True)
def rename(client, old, new, force):
    """Rename the workflow named <old> to <new>."""
    from renku.core.models.refs import LinkReference

    LinkReference(client=client, name=_ref(old)).rename(_ref(new), force=force)


@workflow.command()
@click.argument("name", metavar="<name>")
@pass_local_client(clean=True, commit=True)
def remove(client, name):
    """Remove the remote named <name>."""
    from renku.core.models.refs import LinkReference

    LinkReference(client=client, name=_ref(name)).delete()


@workflow.command()
@click.option("--revision", default="HEAD")
@click.option(
    "-o",
    "--output-file",
    metavar="FILE",
    type=click.Path(exists=False),
    default=None,
    help="Write workflow to the FILE.",
)
@click.argument("paths", type=click.Path(dir_okay=True), nargs=-1)
@pass_local_client
def create(client, output_file, revision, paths):
    """Create a workflow description for a file."""
    graph = Graph(client)
    outputs = graph.build(paths=paths, revision=revision)

    workflow = graph.as_workflow(outputs=outputs,)

    if output_file:
        output_file = Path(output_file)

    wf, path = CWLConverter.convert(workflow, client, path=output_file)

    if not output_file:
        click.echo(wf.export_string())



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service templates create project controller tests."""
import pytest
from marshmallow import ValidationError

from renku.core.utils.scm import normalize_to_ascii


def test_template_create_project_ctrl(ctrl_init, svc_client_templates_creation, mocker):
    """Test template create project controller."""
    from renku.service.controllers.templates_create_project import TemplatesCreateProjectCtrl

    cache, user_data = ctrl_init
    svc_client, headers, payload, rm_remote = svc_client_templates_creation

    ctrl = TemplatesCreateProjectCtrl(cache, user_data, payload)
    ctrl_mock = mocker.patch.object(ctrl, "new_project_push", return_value=None)
    response = ctrl.to_response()

    # Check response.
    assert {"result"} == response.json.keys()
    assert {"url", "namespace", "name"} == response.json["result"].keys()

    # Check ctrl_mock.
    assert ctrl_mock.call_count == 1
    assert response.json["result"]["name"] == ctrl_mock.call_args[0][0].name

    # Ctrl state.
    expected_context = {
        "timestamp",
        "owner",
        "project_namespace",
        "token",
        "email",
        "project_repository",
        "url",
        "identifier",
        "initialized",
        "parameters",
        "project_name",
        "name",
        "new_project_url",
        "fullname",
        "project_slug",
        "git_url",
        "project_name_stripped",
        "depth",
        "ref",
        "new_project_url_with_auth",
        "url_with_auth",
        "user_id",
    }
    assert expected_context.issubset(set(ctrl.context.keys()))

    received_metadata = ctrl.default_metadata
    expected_metadata = {
        "__template_source__",
        "__template_ref__",
        "__template_id__",
        "__namespace__",
        "__automated_update__",
        "__repository__",
        "__sanitized_project_name__",
        "__project_slug__",
    }
    assert expected_metadata == set(received_metadata.keys())

    assert payload["url"] == received_metadata["__template_source__"]
    assert payload["ref"] == received_metadata["__template_ref__"]
    assert payload["identifier"] == received_metadata["__template_id__"]
    assert payload["project_namespace"] == received_metadata["__namespace__"]
    assert payload["project_repository"] == received_metadata["__repository__"]

    assert ctrl.template_version

    project_name = normalize_to_ascii(payload["project_name"])
    assert project_name == received_metadata["__sanitized_project_name__"]
    assert f"{payload['project_namespace']}/{project_name}" == received_metadata["__project_slug__"]


@pytest.mark.parametrize(
    "project_name,expected_name",
    [
        ("Test   renku-core   /é", "test-renku-core"),
        ("Test renku-core é", "test-renku-core"),
        ("Test é renku-core ", "test-renku-core"),
        ("é Test é renku-core ", "test-renku-core"),
        ("Test/renku-core", "test-renku-core"),
        ("Test 😁", "test"),
        ("invalid wörd", "invalid-w-rd"),
        ("invalid wörd and another invalid wórd", "invalid-w-rd-and-another-invalid-w-rd"),
        ("João", "jo-o"),
        ("'My Input String", "my-input-string"),
        ("My Input String", "my-input-string"),
        (" a new project ", "a-new-project"),
        ("test!_pro-ject~", "test-pro-ject"),
        ("test!!!!_pro-ject~", "test-pro-ject"),
        ("Test:-)", "test"),
        ("-Test:-)-", "test"),
        ("test----aua", "test-aua"),
        ("test --üäüaua", "test-aua"),
        ("---- test --üäüaua ----", "test-aua"),
        ("---- test --üäü", "test"),
        ("Caffè", "caff"),
        ("my.repo", "my-repo"),
        ("my......repo", "my-repo"),
        ("my_repo", "my-repo"),
        ("my_______repo", "my-repo"),
        ("-.my___repo.", "my-repo"),
        (".my___-...repo..", "my-repo"),
        ("-.-my-repo.", "my-repo"),
    ],
)
def test_project_name_handler(project_name, expected_name, ctrl_init, svc_client_templates_creation, mocker):
    """Test template create project controller correct set of project name."""
    from renku.service.controllers.templates_create_project import TemplatesCreateProjectCtrl

    cache, user_data = ctrl_init
    svc_client, headers, payload, rm_remote = svc_client_templates_creation
    payload["project_name"] = project_name

    ctrl = TemplatesCreateProjectCtrl(cache, user_data, payload)
    mocker.patch.object(ctrl, "new_project_push", return_value=None)
    response = ctrl.to_response()

    # Check response.
    assert {"result"} == response.json.keys()
    assert {"url", "namespace", "name"} == response.json["result"].keys()
    assert expected_name == response.json["result"]["name"]


@pytest.mark.parametrize("project_name", ["здрасти", "---- --üäü ----", "-.-", "...", "----", "~.---", "`~~"])
def test_except_project_name_handler(project_name, ctrl_init, svc_client_templates_creation, mocker):
    """Test template create project controller exception raised."""
    from renku.service.controllers.templates_create_project import TemplatesCreateProjectCtrl

    cache, user_data = ctrl_init
    svc_client, headers, payload, rm_remote = svc_client_templates_creation
    payload["project_name"] = project_name

    with pytest.raises(ValidationError) as exc_info:
        TemplatesCreateProjectCtrl(cache, user_data, payload)

    assert "Invalid `git_url`" in str(exc_info.value)



# -*- coding: utf-8 -*-
#
# Copyright 2018-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Represent the Common Workflow Language types."""

import os
from pathlib import Path

import attr


class _PathFormatterMixin:
    """Format path property."""

    def __str__(self):
        """Simple conversion to string."""
        return os.path.relpath(os.path.abspath(str(self.path)), os.path.realpath(os.getcwd()))


@attr.s
class File(_PathFormatterMixin):
    """Represent a file."""

    path = attr.ib(converter=Path)


@attr.s
class Directory(_PathFormatterMixin):
    """Represent a directory."""

    # TODO add validation to allow only directories
    path = attr.ib(default=None)
    listing = attr.ib(default=attr.Factory(list))


PATH_OBJECTS = {"File", "Directory"}
PATH_TYPES = (File, Directory)


@attr.s
class Dirent(object):
    """Define a file or subdirectory."""

    entryname = attr.ib(default=None)
    entry = attr.ib(default=None)
    writable = attr.ib(default=False)



# -*- coding: utf-8 -*-
#
# Copyright 2017, 2018 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Create an empty Renku project or reinitialize an existing one.

Start a Renku project
~~~~~~~~~~~~~~~~~~~~~

If you have an existing directory which you want to turn into a Renku project,
you can type:

.. code-block:: console

    $ cd ~/my_project
    $ renku init

or:

.. code-block:: console

    $ renku init ~/my_project

This creates a new subdirectory named ``.renku`` that contains all the
necessary files for managing the project configuration.

If provided directory does not exist, it will be created.

Use a different template
~~~~~~~~~~~~~~~~~~~~~~~~

Renku is installed together with a specific set of templates you can select
when you initialize a project. You can check them by typing:

.. code-block:: console

    $ renku init --list-templates

    INDEX ID     DESCRIPTION                     PARAMETERS
    ----- ------ ------------------------------- -----------------------------
    1     python The simplest Python-based [...] description: project des[...]
    2     R      R-based renku project with[...] description: project des[...]

If you know which template you are going to use, you can provide either the id
``--template-id`` or the template index number ``--template-index``.

You can use a newer version of the templates or even create your own one and
provide it to the ``init`` command by specifying the target template repository
source ``--template-source`` (both local path and remote url are supported) and
the reference ``--template-ref`` (branch, tag or commit).

You can take inspiration from the
`official Renku template repository
<https://github.com/SwissDataScienceCenter/renku-project-template>`_

.. code-block:: console

    $ renku init --template-ref master --template-source \
    https://github.com/SwissDataScienceCenter/renku-project-template

    Fetching template from
    https://github.com/SwissDataScienceCenter/renku-project-template@master
    ... OK

    INDEX ID             DESCRIPTION                PARAMETERS
    ----- -------------- -------------------------- ----------------------
    1     python-minimal Basic Python Project:[...] description: proj[...]
    2     R-minimal      Basic R Project: The [...] description: proj[...]

    Please choose a template by typing the index:

Provide parameters
~~~~~~~~~~~~~~~~~-

Some templates require parameters to properly initialize a new project. You
can check them by listing the templates ``--list-templates``.

To provide parameters, use the ``--parameter`` option and provide each
parameter using ``--parameter "param1"="value1"``.

.. code-block:: console

    $ renku init --template-id python-minimal --parameter \
    "description"="my new shiny project"

    Initializing new Renku repository... OK

If you don't provide the required parameters through the option
``-parameter``, you will be asked to provide them. Empty values are allowed
and passed to the template initialization function.

.. note:: Every project requires a ``name`` that can either be provided using
   ``--name`` or automatically taken from the target folder. This is
   also considered as a special parameter, therefore it's automatically added
   to the list of parameters forwarded to the ``init`` command.

Update an existing project
~~~~~~~~~~~~~~~~~~~~~~~~~~

There are situations when the required structure of a Renku project needs
to be recreated or you have an **existing** Git repository. You can solve
these situation by simply adding the ``--force`` option.

.. code-block:: console

    $ git init .
    $ echo "# Example\nThis is a README." > README.md
    $ git add README.md
    $ git commit -m 'Example readme file'
    # renku init would fail because there is a git repository
    $ renku init --force

You can also enable the external storage system for output files, if it
was not installed previously.

.. code-block:: console

    $ renku init --force --external-storage

"""

import configparser
import os
from collections import OrderedDict, namedtuple
from pathlib import Path
from tempfile import mkdtemp

import attr
import click
from git import Repo

from renku.core import errors
from renku.core.commands.client import pass_local_client
from renku.core.commands.echo import INFO
from renku.core.commands.git import set_git_home
from renku.core.commands.init import create_from_template, fetch_template
from renku.core.commands.options import option_external_storage_requested
from renku.core.models.tabulate import tabulate

_GITLAB_CI = ".gitlab-ci.yml"
_DOCKERFILE = "Dockerfile"
_REQUIREMENTS = "requirements.txt"
CI_TEMPLATES = [_GITLAB_CI, _DOCKERFILE, _REQUIREMENTS]

INVALID_DATA_DIRS = [".", ".renku", ".git"]
"""Paths that cannot be used as data directory name."""


def parse_parameters(ctx, param, value):
    """Parse parameters to dictionary."""
    parameters = {}
    for parameter in value:
        splitted = parameter.split("=", 1)
        if len(splitted) < 2 or len(splitted[0]) < 1:
            raise errors.ParameterError(
                'Parameter format must be --parameter "param1"="value". ', f'--parameter "{parameter}"'
            )
        parameters[splitted[0]] = splitted[1]
    return parameters


def validate_name(ctx, param, value):
    """Validate a project name."""
    if not value:
        value = os.path.basename(ctx.params["path"].rstrip(os.path.sep))
    return value


def store_directory(value):
    """Store directory as a new Git home."""
    Path(value).mkdir(parents=True, exist_ok=True)
    set_git_home(value)
    return value


def create_template_sentence(templates, describe=False, instructions=False):
    """Create templates choice sentence.

    :ref templates: list of templates coming from manifest file
    :ref instructions: add instructions
    """
    Template = namedtuple("Template", ["index", "id", "description", "variables"])

    def extract_description(template_elem):
        """Extract description from template manifest."""
        if describe:
            return template_elem["description"]
        return None

    def extract_variables(template_elem):
        """Extract variables from tempalte manifest."""
        if describe:
            return "\n".join(
                [f"{variable[0]}: {variable[1]}" for variable in template_elem.get("variables", {}).items()]
            )

        return ",".join(template_elem.get("variables", {}).keys())

    templates_friendly = [
        Template(
            index=index + 1,
            id=template_elem["folder"],
            description=extract_description(template_elem),
            variables=extract_variables(template_elem),
        )
        for index, template_elem in enumerate(templates)
    ]

    table_headers = OrderedDict((("index", "Index"), ("id", "Id"), ("variables", "Parameters"),))

    if describe:
        table_headers["description"] = "Description"

    text = tabulate(templates_friendly, headers=table_headers)

    if not instructions:
        return text
    return "{0}\nPlease choose a template by typing the index".format(text)


def is_path_empty(path):
    """Check if path contains files.

    :ref path: target path
    """
    gen = Path(path).glob("**/*")
    return not any(gen)


def resolve_data_directory(data_dir, path):
    """Check data directory is within the project path."""
    if not data_dir:
        return

    absolute_data_dir = (Path(path) / data_dir).resolve()

    try:
        data_dir = absolute_data_dir.relative_to(path)
    except ValueError:
        raise errors.ParameterError(f"Data directory {data_dir} is not within project {path}")

    if str(data_dir).rstrip(os.path.sep) in INVALID_DATA_DIRS:
        raise errors.ParameterError(f"Cannot use {data_dir} as data directory.")

    return data_dir


def check_git_user_config():
    """Check that git user information is configured."""
    dummy_git_folder = mkdtemp()
    repo = Repo.init(dummy_git_folder)
    git_config = repo.config_reader()
    try:
        git_config.get_value("user", "name", None)
        git_config.get_value("user", "email", None)
        return True
    except (configparser.NoOptionError, configparser.NoSectionError):
        return False


@click.command()
@click.argument(
    "path", default=".", type=click.Path(writable=True, file_okay=False, resolve_path=True),
)
@click.option(
    "-n", "--name", callback=validate_name, help="Provide a custom project name.",
)
@click.option(
    "--data-dir",
    default=None,
    type=click.Path(writable=True, file_okay=False),
    help="Data directory within the project",
)
@click.option("-t", "--template-id", help="Provide the id of the template to use.")
@click.option(
    "-i", "--template-index", help="Provide the index number of the template to use.", type=int,
)
@click.option("-s", "--template-source", help="Provide the templates repository url or path.")
@click.option(
    "-r", "--template-ref", default="master", help="Specify the reference to checkout on remote template repository.",
)
@click.option(
    "-p",
    "--parameter",
    "metadata",
    multiple=True,
    type=click.STRING,
    callback=parse_parameters,
    help=(
        "Provide parameters value. Should be invoked once per parameter. "
        'Please specify the values as follow: --parameter "param1"="value"'
    ),
)
@click.option("-l", "--list-templates", is_flag=True, help="List templates available in the template-source.")
@click.option("-d", "--describe", is_flag=True, help="Show description for templates and parameters")
@click.option("--force", is_flag=True, help="Override target path.")
@option_external_storage_requested
@pass_local_client
@click.pass_context
def init(
    ctx,
    client,
    external_storage_requested,
    path,
    name,
    template_id,
    template_index,
    template_source,
    template_ref,
    metadata,
    list_templates,
    force,
    describe,
    data_dir,
):
    """Initialize a project in PATH. Default is the current path."""
    # verify dirty path
    if not is_path_empty(path) and not force and not list_templates:
        existing_paths = [str(p.relative_to(path)) for p in Path(path).iterdir()]
        existing_paths.sort()
        raise errors.InvalidFileOperation(
            f'Folder "{str(path)}" is not empty and contains the following files/directories:'
            + "".join((f"\n\t{e}" for e in existing_paths))
            + "\nPlease add --force flag to transform it into a Renku repository."
        )

    data_dir = resolve_data_directory(data_dir, path)

    if not check_git_user_config():
        raise errors.ConfigurationError(
            "The user name and email are not configured. "
            'Please use the "git config" command to configure them.\n\n'
            '\tgit config --global --add user.name "John Doe"\n'
            "\tgit config --global --add user.email "
            '"john.doe@example.com"\n'
        )

    template_manifest, template_folder, template_source, template_version = fetch_template(
        template_source, template_ref, click.echo
    )

    # select specific template
    repeat = False
    template_data = None
    if template_id:
        if template_index:
            raise errors.ParameterError("Use either --template-id or --template-index, not both", '"--template-index"')
        template_filtered = [
            template_elem for template_elem in template_manifest if template_elem["folder"] == template_id
        ]
        if len(template_filtered) == 1:
            template_data = template_filtered[0]
        else:
            click.echo(f'The template with id "{template_id}" is not available.')
            repeat = True

    if template_index or template_index == 0:
        if template_index > 0 and template_index <= len(template_manifest):
            template_data = template_manifest[template_index - 1]
        else:
            click.echo(f"The template at index {template_index} is not available.")
            repeat = True

    if list_templates:
        if template_data:
            click.echo(create_template_sentence([template_data], describe=describe))
        else:
            click.echo(create_template_sentence(template_manifest, describe=describe))
        return

    if repeat or not (template_id or template_index):
        templates = [template_elem for template_elem in template_manifest]
        if len(templates) == 1:
            template_data = templates[0]
        else:
            template_index = click.prompt(
                text=create_template_sentence(templates, describe=describe, instructions=True),
                type=click.IntRange(1, len(templates)),
                show_default=False,
                show_choices=False,
            )
            template_data = templates[template_index - 1]

        template_id = template_data["folder"]

    # verify variables have been passed
    template_variables = template_data.get("variables", {})
    template_variables_keys = set(template_variables.keys())
    input_parameters_keys = set(metadata.keys())
    for key in template_variables_keys - input_parameters_keys:
        value = click.prompt(
            text=(f'The template requires a value for "{key}" ' f"({template_variables[key]})"),
            default="",
            show_default=False,
        )
        metadata[key] = value
    useless_variables = input_parameters_keys - template_variables_keys
    if len(useless_variables) > 0:
        click.echo(
            INFO + "These parameters are not used by the template and were "
            "ignored:\n\t{}".format("\n\t".join(useless_variables))
        )
        for key in useless_variables:
            del metadata[key]

    # set local path and storage
    store_directory(path)
    if not client.external_storage_requested:
        external_storage_requested = False
    ctx.obj = client = attr.evolve(
        client, path=path, data_dir=data_dir, external_storage_requested=external_storage_requested
    )
    if not is_path_empty(path):
        from git import GitCommandError

        try:
            commit = client.find_previous_commit("*")
            branch_name = "pre_renku_init_{0}".format(commit.hexsha[:7])
            with client.worktree(
                path=path,
                branch_name=branch_name,
                commit=commit,
                merge_args=["--no-ff", "-s", "recursive", "-X", "ours", "--allow-unrelated-histories"],
            ):
                click.echo("Saving current data in branch {0}".format(branch_name))
        except AttributeError:
            click.echo("Warning! Overwriting non-empty folder.")
        except GitCommandError as e:
            click.UsageError(e)

    # supply additional metadata
    metadata["__template_source__"] = template_source
    metadata["__template_ref__"] = template_ref
    metadata["__template_id__"] = template_id
    metadata["__namespace__"] = ""
    metadata["__sanitized_project_name__"] = ""
    metadata["__repository__"] = ""
    metadata["__project_slug__"] = ""

    # clone the repo
    template_path = template_folder / template_data["folder"]
    click.echo("Initializing new Renku repository... ", nl=False)
    with client.lock:
        try:
            create_from_template(
                template_path=template_path,
                client=client,
                name=name,
                metadata=metadata,
                template_version=template_version,
                immutable_template_files=template_data.get("immutable_template_files", []),
                automated_update=template_data.get("allow_template_update", False),
                force=force,
                data_dir=data_dir,
            )
        except FileExistsError as e:
            raise click.UsageError(e)

    # Install git hooks
    from .githooks import install

    ctx.invoke(install, force=force)



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helpers utils for interacting with remote source code management tools."""
import re


def is_ascii(data):
    """Check if provided string contains only ascii characters."""
    return len(data) == len(data.encode())


def normalize_to_ascii(input_string, sep="-"):
    """Adjust chars to make the input compatible as scm source."""
    replace_all = [sep, "_", "."]
    for replacement in replace_all:
        input_string = input_string.replace(replacement, " ")

    return (
        sep.join(
            [
                component
                for component in re.sub(r"[^a-zA-Z0-9_.-]+", " ", input_string).split(" ")
                if component and is_ascii(component)
            ]
        )
        .lower()
        .strip(sep)
    )


def git_unicode_unescape(s, encoding="utf-8"):
    """Undoes git/gitpython unicode encoding."""
    if s.startswith('"'):
        return s.strip('"').encode("latin1").decode("unicode-escape").encode("latin1").decode(encoding)
    return s



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service template create project controller."""
import shutil

import git
from marshmallow import EXCLUDE

from renku.core.commands.init import create_from_template_local, read_template_manifest
from renku.core.errors import RenkuException
from renku.core.utils.contexts import chdir
from renku.service.controllers.api.abstract import ServiceCtrl
from renku.service.controllers.api.mixins import ReadOperationMixin
from renku.service.controllers.utils.project_clone import user_project_clone
from renku.service.serializers.templates import ProjectTemplateRequest, ProjectTemplateResponseRPC
from renku.service.utils import make_new_project_path, new_repo_push
from renku.service.views import result_response


class TemplatesCreateProjectCtrl(ServiceCtrl, ReadOperationMixin):
    """Template create project controller."""

    REQUEST_SERIALIZER = ProjectTemplateRequest()
    RESPONSE_SERIALIZER = ProjectTemplateResponseRPC()

    def __init__(self, cache, user_data, request_data):
        """Construct a templates read manifest controller."""
        self.ctx = TemplatesCreateProjectCtrl.REQUEST_SERIALIZER.load({**user_data, **request_data}, unknown=EXCLUDE)
        super(TemplatesCreateProjectCtrl, self).__init__(cache, user_data, request_data)

        self.template = None

    @property
    def context(self):
        """Controller context."""
        return self.ctx

    @property
    def default_metadata(self):
        """Default metadata for project creation."""
        automated_update = True
        if self.template and "allow_template_update" in self.template:
            automated_update = self.template["allow_template_update"]

        metadata = {
            "__template_source__": self.ctx["git_url"],
            "__template_ref__": self.ctx["ref"],
            "__template_id__": self.ctx["identifier"],
            "__namespace__": self.ctx["project_namespace"],
            "__automated_update__": automated_update,
            "__repository__": self.ctx["project_repository"],
            "__sanitized_project_name__": self.ctx["project_name_stripped"],
            "__project_slug__": self.ctx["project_slug"],
        }

        return metadata

    @property
    def git_user(self):
        """Extract git user from the user data."""
        return {
            "email": self.user_data["email"],
            "name": self.user_data["fullname"],
        }

    def setup_new_project(self):
        """Setup new project for initialization."""
        new_project_path = make_new_project_path(self.user_data, self.ctx)

        if new_project_path.exists():
            shutil.rmtree(new_project_path)
        new_project_path.mkdir(parents=True, exist_ok=True)

        return new_project_path

    def setup_template(self):
        """Reads template manifest."""
        project = user_project_clone(self.user_data, self.ctx)
        templates = read_template_manifest(project.abs_path)
        self.template = next((template for template in templates if template["folder"] == self.ctx["identifier"]), None)
        if self.template is None:
            raise RenkuException("invalid identifier for target repository")

        repo = git.Repo(str(project.abs_path))
        self.template_version = repo.head.commit.hexsha

        # Verify missing parameters
        template_parameters = self.template.get("variables", {})
        provided_parameters = {p["key"]: p["value"] for p in self.ctx["parameters"]}
        missing_keys = list(template_parameters.keys() - provided_parameters.keys())
        if len(missing_keys) > 0:
            raise RenkuException(f"missing parameter: {missing_keys[0]}")

        return project, provided_parameters

    def new_project_push(self, project_path):
        """Push new project to the remote."""
        return new_repo_push(project_path, self.ctx["new_project_url_with_auth"])

    def new_project(self):
        """Create new project from template."""
        template_project, provided_parameters = self.setup_template()
        new_project_path = self.setup_new_project()

        source_path = template_project.abs_path / self.ctx["identifier"]
        with chdir(new_project_path):
            create_from_template_local(
                source_path,
                self.ctx["project_name"],
                provided_parameters,
                self.default_metadata,
                self.template_version,
                self.template.get("immutable_template_files", []),
                self.template.get("allow_template_update", False),
                self.git_user,
                self.ctx["url"],
                self.ctx["ref"],
                "service",
            )
        self.new_project_push(new_project_path)

        return {
            "url": self.ctx["new_project_url"],
            "namespace": self.ctx["project_namespace"],
            "name": self.ctx["project_name_stripped"],
        }

    def renku_op(self):
        """Renku operation for the controller."""
        # NOTE: We leave it empty since it does not execute renku operation.
        pass

    def to_response(self):
        """Execute controller flow and serialize to service response."""
        return result_response(TemplatesCreateProjectCtrl.RESPONSE_SERIALIZER, self.new_project())



# -*- coding: utf-8 -*-
#
# Copyright 2019-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test location of indirect inputs/outputs files list."""

import os
from pathlib import Path

import pytest

from renku.core import errors
from renku.core.models.cwl import command_line_tool


@pytest.mark.serial
@pytest.mark.parametrize("env_var", ["", ".", "some/path/", "./path"])
def test_set_indirect_input_files_paths_via_env_var(tmp_path, env_var, reset_environment):
    """Test setting of RENKU_INDIRECT_PATH env variable."""
    os.environ["RENKU_INDIRECT_PATH"] = env_var

    path = command_line_tool.get_indirect_inputs_path(tmp_path)

    assert path.is_absolute()

    expected_path = Path(".renku") / "tmp" / env_var / "inputs.txt"

    assert expected_path == path.relative_to(tmp_path)


@pytest.mark.serial
@pytest.mark.parametrize("env_var", ["", ".", "some/path/", "./path"])
def test_set_indirect_output_files_paths_via_env_var(tmp_path, env_var, reset_environment):
    """Test setting of RENKU_INDIRECT_PATH env variable."""
    os.environ["RENKU_INDIRECT_PATH"] = env_var

    path = command_line_tool.get_indirect_outputs_path(tmp_path)

    assert path.is_absolute()

    expected_path = Path(".renku") / "tmp" / env_var / "outputs.txt"

    assert expected_path == path.relative_to(tmp_path)


@pytest.mark.serial
@pytest.mark.parametrize("env_var", ["/absolute/path", "..", "../../outside/repo"])
def test_set_invalid_values_for_indirect_env_var(tmp_path, env_var, reset_environment):
    """Test setting invalid values for RENKU_INDIRECT_PATH env variable."""
    os.environ["RENKU_INDIRECT_PATH"] = env_var

    with pytest.raises(errors.InvalidFileOperation):
        command_line_tool.get_indirect_inputs_path(tmp_path)



# -*- coding: utf-8 -*-
#
# Copyright 2017-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Get and set Renku repository or global options.

Set values
~~~~~~~~~~

You can set various Renku configuration options, for example the image registry
URL, with a command like:

.. code-block:: console

    $ renku config set registry https://registry.gitlab.com/demo/demo

By default, configuration is stored locally in the project's directory. Use
``--global`` option to store configuration for all projects in your home
directory.

Remove values
~~~~~~~~~~~~~

To remove a specific key from configuration use:

.. code-block:: console

    $ renku config remove registry

By default, only local configuration is searched for removal. Use ``--global``
option to remove a global configuration value.

Query values
~~~~~~~~~~~~

You can display all configuration values with:

.. code-block:: console

    $ renku config show

Both local and global configuration files are read. Values in local
configuration take precedence over global values. Use ``--local`` or
``--global`` flag to read corresponding configuration only.

You can provide a KEY to display only its value:

.. code-block:: console

    $ renku config show registry
    https://registry.gitlab.com/demo/demo

Available configuration values
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

The following values are available for the ``renku config`` command:

+------------------------+-------------------------------------+-----------+
| Name                   | Description                         | Default   |
+========================+=====================================+===========+
| registry               | The image registry to store Docker  | ``None``  |
|                        | images in                           |           |
+------------------------+-------------------------------------+-----------+
| zenodo.access_token    | Access token for Zenodo API         | ``None``  |
+------------------------+-------------------------------------+-----------+
| dataverse.access_token | Access token for Dataverse API      | ``None``  |
+------------------------+-------------------------------------+-----------+
| dataverse.server_url   | URL for the Dataverse API server    | ``None``  |
|                        | to use                              |           |
+------------------------+-------------------------------------+-----------+
| show_lfs_message       | Whether to show messages about      | ``True``  |
|                        | files being added to git LFS or not |           |
+------------------------+-------------------------------------+-----------+
| lfs_threshold          | Threshold file size below which     | ``100kb`` |
|                        | files are not added to git LFS      |           |
+------------------------+-------------------------------------+-----------+
"""
import click

from renku.core import errors
from renku.core.commands.config import read_config, update_config


@click.group()
def config():
    """Interact with renku configuration."""
    pass


@config.command()
@click.argument("key", required=False, default=None)
@click.option("--local", "local_only", is_flag=True, help="Read from local configuration only.")
@click.option("--global", "global_only", is_flag=True, help="Read from global configuration only.")
def show(key, local_only, global_only):
    """Show current configuration.

    KEY is of the form <group>.<entry>, e.g. 'interactive.default_url'.
    """
    if local_only and global_only:
        raise errors.UsageError("Cannot use --local and --global together.")

    value = read_config(key, local_only, global_only)
    click.secho(value)


@config.command("set")
@click.argument("key")
@click.argument("value")
@click.option("--global", "global_only", is_flag=True, help="Store to global configuration only.")
def set_(key, value, global_only):
    """Set a configuration value.

    KEY is of the form <group>.<entry>, e.g. 'interactive.default_url'.
    """
    update_config(key, value=value, global_only=global_only)
    click.secho("OK", fg="green")


@config.command()
@click.argument("key")
@click.option("--global", "global_only", is_flag=True, help="Remove from global configuration only.")
def remove(key, global_only):
    """Remove a configuration value.

    KEY is of the form <group>.<entry>, e.g. 'interactive.default_url'.
    """
    update_config(key, remove=True, global_only=global_only)
    click.secho("OK", fg="green")



# -*- coding: utf-8 -*-
#
# Copyright 2019-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test behavior of indirect inputs/outputs files list."""

from pathlib import Path

from renku.core.models.entities import Collection, Entity
from renku.core.utils.contexts import chdir


def test_indirect_inputs_outputs(renku_cli, client):
    """Test indirect inputs/outputs that are programmatically created."""
    with chdir(client.path):
        Path("foo").mkdir()
        Path(".renku/tmp").mkdir()
        Path("foo/bar").touch()
        Path("baz").touch()
        Path("qux").touch()

        Path("script.sh").write_text(
            """
            echo foo >> .renku/tmp/inputs.txt
            echo baz >> .renku/tmp/inputs.txt
            echo qux >> .renku/tmp/outputs.txt
            """
        )

        client.repo.git.add("--all")
        client.repo.index.commit("test setup")

    exit_code, plan = renku_cli("run", "sh", "-c", "sh script.sh")

    assert 0 == exit_code
    assert 2 == len(plan.inputs)
    assert 1 == len(plan.arguments)
    plan.inputs.sort(key=lambda e: e.consumes.path)
    assert "baz" == str(plan.inputs[0].consumes.path)
    assert isinstance(plan.inputs[0].consumes, Entity)
    assert plan.inputs[0].position is None
    assert "foo" == str(plan.inputs[1].consumes.path)
    assert isinstance(plan.inputs[1].consumes, Collection)
    assert plan.inputs[1].position is None

    assert 1 == len(plan.outputs)
    assert "qux" == plan.outputs[0].produces.path


def test_duplicate_indirect_inputs(renku_cli, client):
    """Test duplicate indirect inputs are only included once."""
    with chdir(client.path):
        Path("foo").mkdir()
        Path(".renku/tmp").mkdir()
        Path("foo/bar").touch()
        Path("baz").touch()

        Path("script.sh").write_text(
            """
            echo foo/bar >> .renku/tmp/inputs.txt
            echo foo/bar >> .renku/tmp/inputs.txt
            echo foo/./bar >> .renku/tmp/inputs.txt
            echo foo/../foo/bar >> .renku/tmp/inputs.txt
            echo baz >> .renku/tmp/inputs.txt
            echo baz >> .renku/tmp/inputs.txt
            """
        )

        client.repo.git.add("--all")
        client.repo.index.commit("test setup")

    exit_code, plan = renku_cli("run", "--no-output", "sh", "-c", "sh script.sh", "baz")

    assert 0 == exit_code
    assert {"baz", "foo/bar"} == {i.consumes.path for i in plan.inputs}


def test_duplicate_indirect_outputs(renku_cli, client):
    """Test duplicate indirect outputs are only included once."""
    with chdir(client.path):
        Path("foo").mkdir()
        Path(".renku/tmp").mkdir()
        Path("foo/bar").touch()

        Path("script.sh").write_text(
            """
            echo foo/bar >> .renku/tmp/outputs.txt
            echo foo/bar >> .renku/tmp/outputs.txt
            echo foo/./bar >> .renku/tmp/outputs.txt
            echo foo/../foo/bar >> .renku/tmp/outputs.txt
            echo baz >> .renku/tmp/outputs.txt
            echo baz >> .renku/tmp/outputs.txt

            touch baz
            """
        )

        client.repo.git.add("--all")
        client.repo.index.commit("test setup")

    exit_code, plan = renku_cli("run", "sh", "-c", "sh script.sh")

    assert 0 == exit_code
    assert {"baz", "foo/bar"} == {o.produces.path for o in plan.outputs}


def test_indirect_parameters(renku_cli, client):
    """Test indirect parameters."""
    with chdir(client.path):
        Path(".renku/tmp").mkdir()

        Path("script.sh").write_text(
            """
            echo param 1: 42 >> .renku/tmp/parameters.yml
            echo param-2: 42.42 >> .renku/tmp/parameters.yml
            echo param3: 42 >> .renku/tmp/parameters.yml
            echo param 1: "forty-two" >> .renku/tmp/parameters.yml
            """
        )

        client.repo.git.add("--all")
        client.repo.index.commit("test setup")

    exit_code, plan = renku_cli("run", "--no-output", "sh", "-c", "sh script.sh")

    assert 0 == exit_code
    assert {"param 1", "param-2", "param3"} == {a.name for a in plan.run_parameters}
    assert {"forty-two", "42.42", "42"} == {a.value for a in plan.run_parameters}
    assert {"str", "float", "int"} == {a.type for a in plan.run_parameters}

    param_1 = next(p for p in plan.run_parameters if p.name == "param 1")
    assert " " not in param_1._id


def test_indirect_parameters_update(renku_cli, client):
    """Test updating of indirect parameters."""
    with chdir(client.path):
        Path(".renku/tmp").mkdir(exist_ok=True)

        Path("script.sh").write_text(
            """
            echo param 1: "forty-two" >> .renku/tmp/parameters.yml
            echo param-2: 42.42 >> .renku/tmp/parameters.yml
            echo param3: 42 >> .renku/tmp/parameters.yml
            """
        )

        client.repo.git.add("--all")
        client.repo.index.commit("test setup")

    renku_cli("run", "sh", "script.sh", stdout="result")

    with chdir(client.path):
        Path(".renku/tmp").mkdir(exist_ok=True)

        Path("script.sh").write_text(
            """
            echo param 1: "forty-two-updated" >> .renku/tmp/parameters.yml
            echo param-2: 42.42 >> .renku/tmp/parameters.yml
            """
        )

        client.repo.git.add("--all")
        client.repo.index.commit("test setup")

    exit_code, plan = renku_cli("update", "--all")

    assert 0 == exit_code
    assert {"forty-two-updated", "42.42", "42"} == {a.value for a in plan.run_parameters}



# -*- coding: utf-8 -*-
#
# Copyright 2019-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test behavior of ``--output`` option."""

import os
from pathlib import Path

from renku.core.models.entities import Collection


def test_run_succeeds_normally(renku_cli, client, subdirectory):
    """Test when an output is detected"""
    foo = os.path.relpath(client.path / "foo", os.getcwd())
    exit_code, cwl = renku_cli("run", "touch", foo)

    assert 0 == exit_code
    assert 0 == len(cwl.inputs)
    assert 1 == len(cwl.outputs)
    assert "foo" == cwl.outputs[0].produces.path


def test_when_no_change_in_outputs_is_detected(renku_cli, subdirectory):
    """Test when no output is detected"""
    renku_cli("run", "touch", "foo")
    exit_code, _ = renku_cli("run", "ls", "foo")

    assert 1 == exit_code


def test_with_no_output_option(renku_cli, client, subdirectory):
    """Test --no-output option with no output detection"""
    foo = os.path.relpath(client.path / "foo", os.getcwd())
    renku_cli("run", "touch", foo)
    exit_code, cwl = renku_cli("run", "--no-output", "touch", foo)

    assert 0 == exit_code
    assert 1 == len(cwl.inputs)
    assert "foo" == str(cwl.inputs[0].consumes.path)
    assert 0 == len(cwl.outputs)


def test_explicit_outputs_and_normal_outputs(renku_cli, client, subdirectory):
    """Test explicit outputs and normal outputs can co-exist"""
    foo = os.path.relpath(client.path / "foo", os.getcwd())
    os.mkdir(foo)
    bar = os.path.relpath(client.path / "bar", os.getcwd())
    renku_cli("run", "touch", bar)
    baz = os.path.relpath(client.path / "baz", os.getcwd())
    qux = os.path.join(foo, "qux")

    exit_code, cwl = renku_cli("run", "--output", foo, "--output", bar, "touch", baz, qux)

    assert 0 == exit_code
    cwl.inputs.sort(key=lambda e: e.position)
    assert 4 == len(cwl.outputs)
    assert {"foo", "bar", "baz", "foo/qux"} == {str(o.produces.path) for o in cwl.outputs}


def test_explicit_outputs_and_std_output_streams(renku_cli, client, subdirectory):
    """Test that unchanged std output streams can be marked with explicit
    outputs"""
    exit_code, _ = renku_cli("run", "echo", "foo", stdout="bar")
    assert 0 == exit_code

    exit_code, _ = renku_cli("run", "echo", "foo", stdout="bar")
    assert 1 == exit_code

    exit_code, _ = renku_cli("run", "--output", "bar", "echo", "foo", stdout="bar")
    assert 0 == exit_code


def test_output_directory_with_output_option(renku_cli, client, subdirectory):
    """Test output directories are not deleted with --output"""
    outdir = os.path.relpath(client.path / "outdir", os.getcwd())
    a_script = ("sh", "-c", 'mkdir -p "$0"; touch "$0/$1"')
    renku_cli("run", *a_script, outdir, "foo")

    exit_code, _ = renku_cli("run", "--output", outdir, *a_script, outdir, "bar")

    assert 0 == exit_code
    assert (client.path / "outdir" / "foo").exists()
    assert (client.path / "outdir" / "bar").exists()


def test_output_directory_without_separate_outputs(renku_cli, client):
    """Test output files not listed as separate outputs.

    See https://github.com/SwissDataScienceCenter/renku-python/issues/387
    """
    a_script = ("sh", "-c", 'mkdir -p "$0"; touch "$0/$1"')
    exit_code, cwl = renku_cli("run", *a_script, "outdir", "foo")

    assert 0 == exit_code
    assert 1 == len(cwl.outputs)
    assert isinstance(cwl.outputs[0].produces, Collection)


def test_explicit_inputs_must_exist(renku_cli):
    """Test explicit inputs exist before run"""
    exit_code, _ = renku_cli("run", "--input", "foo", "touch", "bar")

    assert 2 == exit_code


def test_explicit_inputs_are_inside_repo(renku_cli):
    """Test explicit inputs are inside the Renku repo"""
    exit_code, _ = renku_cli("run", "--input", "/tmp", "touch", "foo")

    assert 2 == exit_code


def test_explicit_outputs_must_exist(renku_cli):
    """Test explicit outputs exist after run"""
    exit_code, _ = renku_cli("run", "--output", "foo", "touch", "bar")

    assert 2 == exit_code


def test_explicit_inputs_and_outputs_are_listed(renku_cli, client):
    """Test explicit inputs and outputs will be in generated CWL file"""
    foo = Path(os.path.relpath(client.path / "foo", os.getcwd()))
    foo.mkdir()
    renku_cli("run", "touch", "foo/file")
    renku_cli("run", "touch", "bar", "baz")

    exit_code, cwl = renku_cli("run", "--input", "foo", "--input", "bar", "--output", "baz", "echo")
    assert 0 == exit_code

    assert 2 == len(cwl.inputs)
    cwl.inputs.sort(key=lambda e: e.consumes.path)

    assert cwl.inputs[0].position is None
    assert "bar" == str(cwl.inputs[0].consumes.path)

    assert cwl.inputs[1].position is None
    assert "foo" == str(cwl.inputs[1].consumes.path)
    assert isinstance(cwl.inputs[1].consumes, Collection)

    assert cwl.outputs[0].position is None
    assert not isinstance(cwl.outputs[0].produces, Collection)
    assert "baz" == cwl.outputs[0].produces.path


def test_explicit_inputs_can_be_in_inputs(renku_cli, client, subdirectory):
    """Test explicit inputs that are in inputs are treated as normal inputs"""
    foo = os.path.relpath(client.path / "foo", os.getcwd())
    renku_cli("run", "touch", foo)

    exit_code, cwl = renku_cli("run", "--input", foo, "--no-output", "ls", foo)

    assert 0 == exit_code
    assert 1 == len(cwl.inputs)

    assert "foo" == str(cwl.inputs[0].consumes.path)
    assert not isinstance(cwl.inputs[0].consumes, Collection)

    assert cwl.inputs[0].position is not None


def test_explicit_inputs_in_subdirectories(renku_cli, client):
    """Test explicit inputs that are in sub-dirs are made accessible"""

    # Set up a script with hard dependency
    renku_cli("run", "--no-output", "mkdir", "foo")
    renku_cli("run", "echo", "some changes", stdout="foo/bar")
    renku_cli("run", "echo", "cat foo/bar", stdout="script.sh")

    exit_code, _ = renku_cli("run", "--input", "foo/bar", "--input", "script.sh", "sh", "script.sh", stdout="output")
    assert 0 == exit_code

    # Status must be dirty if foo/bar changes
    renku_cli("run", "echo", "new changes", stdout="foo/bar")
    exit_code, _ = renku_cli("status")
    assert 1 == exit_code

    exit_code, _ = renku_cli("update", "--all")
    assert 0 == exit_code
    assert (client.path / "foo" / "bar").exists()
    assert (client.path / "script.sh").exists()
    assert (client.path / "output").exists()


def test_no_explicit_or_detected_output(renku_cli):
    """Test output detection is disbaled and no explicit output is passed."""
    exit_code, _ = renku_cli("run", "--no-output-detection", "echo")

    assert 1 == exit_code


def test_no_output_and_disabled_detection(renku_cli):
    """Test --no-output works with no output detection."""
    exit_code, cwl = renku_cli("run", "--no-output-detection", "--no-output", "echo")

    assert 0 == exit_code
    assert 0 == len(cwl.inputs)
    assert 0 == len(cwl.outputs)


def test_disabled_detection(renku_cli):
    """Test disabled auto-detection of inputs and outputs."""
    exit_code, cwl = renku_cli(
        "run", "--no-input-detection", "--no-output-detection", "--output", "README.md", "touch", "some-files"
    )

    assert 0 == exit_code
    assert 0 == len(cwl.inputs)
    assert 1 == len(cwl.outputs)
    assert "README.md" == str(cwl.outputs[0].produces.path)


def test_inputs_must_be_passed_with_no_detection(renku_cli, client):
    """Test when detection is disabled, inputs must be explicitly passed."""
    exit_code, cwl = renku_cli(
        "run", "--no-input-detection", "--input", "Dockerfile", "--no-output", "ls", "-l", "README.md", "Dockerfile"
    )

    assert 0 == exit_code

    assert 1 == len(cwl.inputs)
    assert cwl.inputs[0].position is not None
    assert "Dockerfile" == str(cwl.inputs[0].consumes.path)


def test_overlapping_explicit_outputs(renku_cli, client):
    """Test explicit outputs are not removed even if they overlap."""
    foo = Path(os.path.relpath(client.path / "foo", os.getcwd()))
    foo.mkdir()
    renku_cli("run", "touch", "foo/bar")

    exit_code, cwl = renku_cli(
        "run", "--no-input-detection", "--no-output-detection", "--output", "foo", "--output", "foo/bar", "echo"
    )

    assert 0 == exit_code
    assert 0 == len(cwl.inputs)
    assert 2 == len(cwl.outputs)
    assert {"foo", "foo/bar"} == {str(o.produces.path) for o in cwl.outputs}


def test_std_streams_must_be_in_explicits(renku_cli):
    """Test when auto-detection is disabled, std streams must be passed
    explicitly."""
    exit_code, cwl = renku_cli(
        "run", "--no-output-detection", "--output", "Dockerfile", "ls", stdin="README.md", stdout="out", stderr="err"
    )

    assert 0 == exit_code
    assert 1 == len(cwl.inputs)
    assert "README.md" == str(cwl.inputs[0].consumes.path)
    assert 1 == len(cwl.outputs)
    assert "Dockerfile" == str(cwl.outputs[0].produces.path)

    exit_code, cwl = renku_cli(
        "run",
        "--no-input-detection",
        "--no-output-detection",
        "--input",
        "README.md",
        "--output",
        "out",
        "--output",
        "err",
        "ls",
        stdin="Dockerfile",
        stdout="out",
        stderr="err",
    )

    assert 0 == exit_code
    assert 1 == len(cwl.inputs)
    assert "README.md" == str(cwl.inputs[0].consumes.path)
    assert 2 == len(cwl.outputs)
    assert {"out", "err"} == {str(o.produces.path) for o in cwl.outputs}


def test_explicit_input_as_out_streams(renku_cli):
    """Test cannot use explicit inputs as stdout/stderr when auto-detection is
    disabled."""
    exit_code, cwl = renku_cli(
        "run",
        "--no-input-detection",
        "--no-output-detection",
        "--input",
        "README.md",
        "ls",
        stdout="README.md",
        stderr="README.md",
    )

    assert 2 == exit_code


def test_explicit_output_as_stdin(renku_cli):
    """Test cannot use explicit outputs as stdin when auto-detection is
    disabled."""
    exit_code, cwl = renku_cli(
        "run", "--no-input-detection", "--no-output-detection", "--output", "README.md", "ls", stdin="README.md",
    )

    assert 2 == exit_code



# -*- coding: utf-8 -*-
#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Check for large files in Git history."""

from renku.core.commands.echo import WARNING


def check_lfs_info(client):
    """Checks if files in history should be in LFS."""
    if not client.check_external_storage():
        return True, None

    files = client.check_lfs_migrate_info()

    if not files:
        return True, None

    message = (
        WARNING
        + "Git history contains large files - consider moving them "
        + "to external storage like git LFS\n\t"
        + "\n\t".join(files)
        + "\n"
    )

    return False, message
